# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import shutil
import subprocess
import fixtures

from ubuntu_system_tests.common.utils import delete_file
from ubuntu_system_tests.helpers.processes import stop_process
from ubuntu_system_tests.helpers.system_settings import (
    get_default_sim_for_calls, set_default_sim_for_calls)


class DialerTestEnvironment(fixtures.Fixture):
    """Fixture used to prepare the phone environment"""

    def setUp(self):
        super().setUp()
        self.useFixture(SetDefaultSIM())
        self.useFixture(UseEmptyHistory())


class SetDefaultSIM(fixtures.Fixture):
    """
    Fixture to set the default sim for calls.

    """

    def __init__(self, sim='1'):
        """
        Specify what sim to set.

        :param sim: '1', '2' or 'ask'

        """
        super().__init__()
        self.sim = sim

    def setUp(self):
        super().setUp()
        original_value = get_default_sim_for_calls()
        self.addCleanup(set_default_sim_for_calls, original_value)
        set_default_sim_for_calls(self.sim)


class FillCustomHistory(fixtures.Fixture):

    history_db = "dialer_history.sqlite"
    data_sys = "/usr/lib/python3/dist-packages/ubuntu_system_tests/tests/data/databases"  # NOQA
    data_local = "tests/data/databases"
    database_path = os.path.join('/tmp/', history_db)

    prefilled_history_local = os.path.join(data_local, history_db)
    prefilled_history_system = os.path.join(data_sys, history_db)

    def setUp(self):
        super(FillCustomHistory, self).setUp()
        self.addCleanup(self._clear_test_data)
        self.addCleanup(self._kill_service_to_respawn)
        self._clear_test_data()
        self._prepare_history_data()
        self._kill_service_to_respawn()
        self._start_service_with_custom_data()

    def _prepare_history_data(self):
        if os.path.exists(self.prefilled_history_local):
            shutil.copy(self.prefilled_history_local, self.database_path)
        else:
            shutil.copy(self.prefilled_history_system, self.database_path)

    def _clear_test_data(self):
        delete_file(self.database_path)

    def _kill_service_to_respawn(self):
        stop_process('history-daemon')

    def _start_service_with_custom_data(self):
        os.environ['HISTORY_SQLITE_DBPATH'] = self.database_path
        subprocess.Popen(['history-daemon'], stdout=subprocess.DEVNULL,
                         stderr=subprocess.DEVNULL)


class UseEmptyHistory(FillCustomHistory):
    database_path = ':memory:'

    def setUp(self):
        super(UseEmptyHistory, self).setUp()

    def _prepare_history_data(self):
        # just avoid doing anything
        self.database_path = ':memory:'

    def _clear_test_data(self):
        # don't do anything
        self.database_path = ''
