# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ubuntu_system_tests.helpers.indicators.cpo import (
    IndicatorPage, INDICATOR_PAGE_PATH)
from ubuntu_system_tests.helpers.autopilot import validate_dbus_object
from ubuntu_system_tests.helpers.ubuntuuitools.cpo.switchmenu import SwitchMenu  # NOQA

from ubuntu_system_tests.helpers.ubuntuuitools.cpo.common import CustomProxyObjectBase  # NOQA
from ubuntu_system_tests.helpers.unity8.utils import UNITY8_PATH_ROOT

ICON_NAME_PAUSE = 'media-playback-pause'
ICON_NAME_PLAY = 'media-playback-start'


class SoundIndicatorPage(IndicatorPage):
    """Autopilot helper for the sound indicator page."""

    name = 'indicator-sound'

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, UNITY8_PATH_ROOT, INDICATOR_PAGE_PATH,
            objectName='indicator-sound-page')

    @property
    def playback_controls(self):
        """Return the playback controls object."""
        return self.wait_select_single('PlaybackItemMenu')

    def _get_sound_settings_menu_item(self):
        """Return the sound settings list item."""
        return self.select_single(
            objectName='indicator.phone-settings', visible=True)

    def get_silent_mode_switcher(self):
        menu = self.select_single(
            SwitchMenu, objectName='indicator.silent-mode', visible=True)
        return menu.get_checkbox()

    def toggle_silent_mode_state(self):
        """Change the silent mode setting from it's current state and return
        its new checked state."""
        switcher = self.get_silent_mode_switcher()
        switcher.change_state()
        return switcher.checked

    def enable_silent_mode(self):
        """Enable silent mode if it is not already enabled."""
        switcher = self.get_silent_mode_switcher()
        if not switcher.checked:
            switcher.change_state()

    def disable_silent_mode(self):
        """Disable silent mode if it is already enabled."""
        switcher = self.get_silent_mode_switcher()
        if switcher.checked:
            switcher.change_state()

    def open_sound_settings(self):
        """Tap the sound settings list item to open settings."""
        self.pointing_device.click_object(self._get_sound_settings_menu_item())


class PlaybackItemMenu(CustomProxyObjectBase):
    """Autopilot helpers for the MPRIS controls in sound indicator."""

    def _get_play_pause_button(self, **kwargs):
        """Return the playback button."""
        return self.wait_select_single(objectName='playButton', **kwargs)

    def _get_previous_button(self):
        """Return the skip previous button."""
        return self.wait_select_single(objectName='previousButton')

    def _get_next_button(self):
        """Return the skip next button."""
        return self.wait_select_single(objectName='nextButton')

    def is_play_button_enabled(self):
        """Return bool representing if the playback button is enabled."""
        return self._get_play_pause_button().enabled

    def is_next_button_enabled(self):
        """Return bool representing if the skip next button is enabled."""
        return self._get_next_button().enabled

    def is_previous_button_enabled(self):
        """Return bool representing if the skip previous button is enabled."""
        return self._get_previous_button().enabled

    def is_track_playing(self):
        play_button = self._get_play_pause_button()
        if play_button.enabled:
            if play_button.iconName == ICON_NAME_PAUSE:
                return True
        return False

    def click_playback_button(self):
        self.pointing_device.click_object(self._get_play_pause_button())

    def play_track(self):
        """Press the playback button, if its already in "play" state."""
        if not self.is_track_playing():
            self.click_playback_button()
            self.playing.wait_for(True)

    def pause_track(self):
        """Press the playback button, if its already in "pause" state."""
        if self.is_track_playing():
            self.click_playback_button()
            self.playing.wait_for(False)

    def skip_next(self):
        """Tap on the skip next button, if there is a song in the queue."""
        if self.canGoNext:
            next_button = self._get_next_button()
            self.pointing_device.click_object(next_button)

    def skip_previous(self):
        """Tap on the skip previous button."""
        if self.canGoPrevious:
            previous_button = self._get_previous_button()
            self.pointing_device.click_object(previous_button)
