# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015, 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from autopilot import exceptions
from autopilot.introspection import is_element

from ubuntu_system_tests.helpers.autopilot import (
    patched_wait_select_single, validate_dbus_object)
from ubuntu_system_tests.helpers.autopilot.list import KEY_ORDER_BY_Y
from ubuntu_system_tests.helpers.scopes.base import (
    GenericScopeView as GenericScopeViewBase)
from ubuntu_system_tests.helpers.ubuntuuitools.cpo.header import DashPageHeader as Dph  # NOQA
from ubuntu_system_tests.helpers.ubuntuuitools.cpo.qquickgridview import QQuickGridView  # NOQA
from ubuntu_system_tests.helpers.unity8.utils import (
    UNITY8_DASH_PATH_ROOT, get_dash)


class GenericScopeView(GenericScopeViewBase):
    """ This class contains generic methods used by the music scope classes """

    HEADER_MY_MUSIC = 'dashSectionHeadermymusic'
    CATEGORY_MY_MUSIC = 'dashCategorymymusic'
    CATEGORY_ARTISTS = 'dashCategoryartists'
    CATEGORY_ALBUMS = 'dashCategoryalbums'

    ONLINE_MUSIC_CATEGORIES = {
        "7digital": {
            "name": 'dashCategory7digital',
            "header": 'dashSectionHeader7digital'
        }, "youtube": {
            "name": 'dashCategoryyoutube',
            "header": 'dashSectionHeaderyoutube'
        }, "soundcloud": {
            "name": 'dashCategorysoundcloud',
            "header": 'dashSectionHeadersoundcloud'
        }
    }

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, UNITY8_DASH_PATH_ROOT, b'GenericScopeView',
            objectName='musicaggregator')

    def _get_displayed_title_labels(self, object_name, section=None):
        """Return all displayed title labels.
        :param object_name: Object name of required label. Either 'titleLabel'
                            or 'subtitleLabel'.
        :param section: Optional section to examine.
        """
        grids = section.select_many('CardGrid', visible=True) if (
            section is not None) else self.select_many(
            'CardGrid', visible=True)

        labels = []
        for grid in grids:
            for label in grid.select_many(objectName=object_name):
                labels.append(label.text)
        return labels

    @staticmethod
    def get_titles_list():
        """
        Retrieve all the titles displayed within the scope view.
        :returns a list of titles that are currently visible
        """
        return get_dash().select_many(
            objectName='titleLabel', visible=True)

    def get_displayed_titles(self, section=None):
        """
        Retrieve all the titles displayed for all the songs.
        The titles displayed could be the artist, album, genre, etc.
        :param section: If not None, the titles will be retrieved from there,
        otherwise will be retrieved from the scope
        """
        return self._get_displayed_title_labels(
            object_name='titleLabel', section=section)

    def get_displayed_subtitles(self, section=None):
        """
        Retrieve all the sub-titles displayed for all the songs.
        The sub-titles displayed could be the artist, album, genre, etc.
        :param section: If not None, the titles will be retrieved from there,
        otherwise will be retrieved from the scope
        """
        return self._get_displayed_title_labels(
            object_name='subtitleLabel', section=section)

    def is_artist(self, artist):
        """ Determines if there is an album with the desired name in the
        current scope
        :param artist: The name of the desired artist
        :return: True is there is an artist with the desired name, False
        otherwise
        """
        try:
            self.get_artist(artist)
            return True
        except exceptions.StateNotFoundError:
            return False

    def get_artist(self, artist):
        """ Retrieve a list of artists with the desired name. """
        return self.wait_select_many(
            ap_result_sort_keys=KEY_ORDER_BY_Y, objectName='subtitleLabel',
            text=artist)[0]

    def select_artist(self, artist):
        """
        Click on the album with the desired name and wait until the new
        screen is loaded
        :return: A category preview object for the new screen displayed
        """
        self.pointing_device.click_object(self.get_artist(artist))
        self.wait_for_processing_to_complete()
        from ubuntu_system_tests.helpers.scopes.music import _artist
        return get_dash().get_current_scope(_artist.GenericScopeView,
                                            objectName='musicaggregator',
                                            visible=True)

    def get_subtitles(self, category):
        """
        Retrieve all the subtitles displayed for all the songs in
        a given category
        :raise StateNotFoundError: When the category section is not displayed
        """
        return self._get_subtitles(category)

    def is_category_displayed(self, category):
        """ Indicates if a category section is being displayed """
        is_element(self.get_category_section, category)

    def select_category(self, header):
        """
        Click on a given category header, wait until the page is loaded
        and return the category preview object
        """
        return self._select_category(header)

    def get_titles(self, category):
        """
        Retrieve all the titles displayed for all the songs in a given category
        :param category: Category name to be used as a filter
            """
        return self._get_titles(category)

    def _get_titles(self, category):
        """
        Retrieve all the titles displayed for all the songs in my music
        """
        category = self.get_category_section(category)
        return self.get_displayed_titles(section=category)

    def _get_subtitles(self, category_name):
        """
        Retrieve all the subtitles displayed for all the songs in a
        given category
        """
        category = self.get_category_section(category_name)
        return self.get_displayed_subtitles(section=category)

    def _get_section_header(self, header_name):
        return self.wait_select_single(objectName=header_name)

    @staticmethod
    def get_category_section(category_name):
        return get_dash().wait_select_single(
            DashCategoryBase, objectName=category_name, visible=True)

    def _select_category(self, header_name):
        section = self._get_section_header(header_name)
        self.pointing_device.click_object(section)
        self.wait_for_processing_to_complete()
        return self if not header_name == self.HEADER_MY_MUSIC else (
            self.get_category_section(self.CATEGORY_ARTISTS))

    def get_online_music_categories(self, ids):
        """
        Retrieves the online music categories for a given set of ids
        :param ids the ids of the online music categories
        :returns a list of online music categories
        """
        categories = []
        for id in ids:
            categories.append(self.ONLINE_MUSIC_CATEGORIES[id])
        return categories


class DashCategoryBase(GenericScopeView):
    """
    This class represents a category of music displayed in the music scope.

    """

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, UNITY8_DASH_PATH_ROOT, b'DashCategoryBase')

    def _get_button_by_title(self, title):
        return self.wait_select_single('UCAbstractButton', title=title)

    def _get_track_list_view(self):
        """Return the track list view that can be swiped up and down."""
        return self.select_single(
            QQuickGridView, objectName='responsiveGridViewGrid')

    def _click_button_by_title(self, title):
        """ Click on the button with the desired title """
        button = self._get_button_by_title(title)
        self._get_track_list_view().swipe_child_into_view(button)
        self.pointing_device.click_object(button)

    def select_track(self, track):
        """ Click on the track with the desired name """
        from ubuntu_system_tests.helpers.scopes.music import _track
        self._click_button_by_title(track)
        return patched_wait_select_single(
            root=self, class_object=_track.Preview, class_name='Preview',
            objectName='preview0', visible=True)

    def select_artist(self, artist):
        """ Click on the artist with the desired name """
        self._click_button_by_title(artist)
        return get_dash().wait_select_single(objectName=self.CATEGORY_ALBUMS)


class DashPageHeader(Dph):
    """
    This class represents the header section that is part of the Dash
    and implement the methods specific to the Music Header.
    """

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, UNITY8_DASH_PATH_ROOT, b'DashPageHeader')

    def _get_search_button(self):
        """Return header search button."""
        return self.select_single(objectName='search_button')

    def _get_options_list(self):
        """Return options list displayed as panel underneath header."""
        temp_scope = get_dash().select_single(objectName='dashTempScopeItem')
        panel = temp_scope.select_single(objectName='peExtraPanel')
        return panel.select_single(objectName='navigation0')

    def select_search(self, options):
        """Press search button and select search options from list.
        :param options: List of options to select from list.
        :return: Displayed category.
        """
        self.pointing_device.click_object(self._get_search_button())
        options_list = self._get_options_list()
        flickable = options_list.select_single('QQuickFlickable')
        for option in options:
            list_item = flickable.select_single(text=option)
            flickable.swipe_child_into_view(list_item)
            self.pointing_device.click_object(list_item)
        scope = get_dash().get_current_scope()
        scope.wait_for_processing_to_complete()
