# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015, 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import re

from autopilot import exceptions
from retrying import retry

from ubuntu_system_tests.helpers.autopilot import (
    patched_wait_select_single, validate_dbus_object)
from ubuntu_system_tests.helpers.autopilot.list import KEY_ORDER_BY_X
from ubuntu_system_tests.helpers import context
from ubuntu_system_tests.helpers.scopes.base import GenericScopeView
from ubuntu_system_tests.helpers.unity8.utils import UNITY8_DASH_PATH_ROOT

NAME_MY_PHOTOS_GRID = \
    'myphotos:declared:first-result:surface-rdr:surfacing:My Photos'


class GenericScopeView(GenericScopeView):
    """This class represent the Photos scope."""

    SEARCH_FORMAT = '%d %m %Y'

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, UNITY8_DASH_PATH_ROOT, b'GenericScopeView',
            objectName='com.canonical.scopes.photos_photos')

    @retry(stop_max_delay=10000,
           retry_on_exception=lambda exception:
           isinstance(exception, exceptions.StateNotFoundError))
    def get_picture_from_carousel(self, pic_path):
        """ Retrieve the specified picture from the carousel.

        Due to different screen resolutions, some images in the carousel may
        have been re-scaled by the thumbnailer. Because of this the source path
        of the picture may change according to which device the test is being
        run on. In order to over-come this, only the ending of the picture
        source path is checked.

        For example an image which has been generated by the thumbnailer would
        be: image://thumbnailer/file:///home/phablet/Pictures/canonical.png

        But an image that has not been through the thumbnailer woud be:
        file:///home/phablet/Pictures/canonical.png

        :param pic_path: Path of the picture.
        :return: The picture object when the picture is in the carousel.
        :raises: StateNotFoundError if required picture cannot be found.
        """
        my_photos_grid = [grid for grid in
                          self.select_many("DashCategoryBase")
                          if grid.objectName.endswith('My Photos')][0]
        filters = {'objectName': 'artShapeLoader', 'visible': True}
        photos = my_photos_grid.select_many(**filters)
        uri_pattern = re.compile(r"(({})([^?/]*))".format(pic_path))
        for photo in photos:
            match = uri_pattern.search(photo.cardArt)
            if match and match.group(0) == pic_path:
                return photo
        raise exceptions.StateNotFoundError(cardArt=pic_path, **filters)

    def get_pictures_from_carousel(self, pic_path):
        """ Retrieve all the pictures from the carousel which match with path
        passed as parameter.

        :param pic_path: Path of the picture or dir which include pictures.
        :return: The list of pictures displayed in the carousel.
        """
        kwargs = {'objectName': 'artImage', 'visible': True}
        pictures = self.select_many(
            'CroppedImageMinimumSourceSize',
            ap_result_sort_keys=KEY_ORDER_BY_X, **kwargs)
        return [picture for picture in pictures if pic_path in picture.source]

    def select_picture_from_carousel(self, img):
        """ Select the picture in the carousel """
        self.pointing_device.click_object(img)

    def get_current_photo_preview(self):
        """
        Retrieve the photo preview object that is currently being displayed
        :return: The PhotoPreview object
        """
        from ubuntu_system_tests.helpers.scopes.photos import _preview
        return patched_wait_select_single(
            root=self, class_object=_preview.Preview, class_name='PreviewView',
            visible=True)

    @retry(stop_max_delay=context.shared.retry_time,
           retry_on_exception=lambda exception:
           isinstance(exception, exceptions.StateNotFoundError))
    def get_my_photos_label(self):
        return self.select_single('UCLabel', text='My Photos')

    def go_to_my_photos(self):
        """ Go to my photos section """
        my_photos = self.get_my_photos_label()
        self.pointing_device.click_object(my_photos)
        return self._get_photos_list()

    @retry(stop_max_delay=context.shared.retry_time,
           retry_on_exception=lambda exception:
           isinstance(exception, exceptions.StateNotFoundError))
    def _get_photos_list(self):
        """ Retrieve the photos list object that is currently being displayed
        :return: The categoryListView object
        """
        from ubuntu_system_tests.helpers.scopes.photos import _list  # NOQA
        label = context.shared.dash.select_single(
            'UCLabel', objectName='header_title_label', text='My Photos',
            visible=True)
        return label.get_parent('ListViewWithPageHeader')

    def get_header_title(self):
        return self.select_single("DashPageHeader", visible=True).title

    def clear_search(self):
        cancel_btn = self.select_single(text='Cancel', visible=True)
        self.pointing_device.click_object(cancel_btn)
        self.wait_for_processing_to_complete()
