# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
import stat
import subprocess

from autopilot.matchers import Eventually
from testtools.matchers import Equals


from ubuntu_system_tests.helpers.processes import get_process_id
from ubuntu_system_tests.helpers.testbed import (
    run_command_with_sudo as sudo, write_to_file_with_sudo)

logger = logging.getLogger(__name__)

SENSOR_OVERRIDE = 'UBUNTU_PLATFORM_API_TEST_OVERRIDE=sensors'


class SensorsBase:

    @staticmethod
    def write_to_fifo(process_name, commands):
        """Write specified commands to the repowerd sensor fifo buffer.
        :param process_name: the name of the process to write the fifo file
        :param commands: List of commands to write."""
        fifo_path = SensorsBase.get_fifo_path(process_name)
        for command in commands:
            echo_cmd = 'echo "{}"'.format(command)
            write_to_file_with_sudo(fifo_path, echo_cmd,
                                    append=True)

    @staticmethod
    def get_fifo_path(process_name):
        """Return path of fake sensor fifo
        :param process_name: the name of the process to get the fifo file path
        :return the path of the fifo file as a string"""
        return '/tmp/sensor-fifo-{}'.format(get_process_id(process_name))

    @staticmethod
    def validate_fifo_ready(process_name):
        """Wait until the given process fifo buffer is ready and validate.
        :param process_name: the name of the process to validate if
        the fifo file is ready
        """
        fifo_path = SensorsBase.get_fifo_path(process_name)
        Eventually(Equals(True)).match(
            lambda: os.path.exists(fifo_path))
        assert (stat.S_ISFIFO(os.stat(fifo_path).st_mode))

    def setup_sensors(self, process_name):
        """Initialize the fake sensors for the given process.
        :param process_name: the name of the process for which
        the sensors will be set up"""
        cmds = self.get_fake_sensors_setup_commands()
        return SensorsBase.write_to_fifo(process_name, cmds)

    @staticmethod
    def get_fake_sensors_setup_commands():
        raise NotImplementedError('Each particular subclass '
                                  'must implement setup commands method')

    @staticmethod
    def are_fake_sensors_enabled_for_process(pid):
        """Return True if fake sensors enabled for specified process."""
        if not os.path.exists('/tmp/sensor-fifo-{}'.format(pid)):
            return False

        try:
            output = sudo('cat /proc/{pid}/environ'.format(pid=pid)).decode()
        except subprocess.CalledProcessError:
            return False
        return output.find(SENSOR_OVERRIDE) >= 0
