# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import signal

from functools import wraps

MSG_TIMEOUT = '{} did not complete within {} seconds.'

logger = logging.getLogger(__name__)


class TimeoutHandler:
    """Timeout handling class that raises an exception after timeout period."""

    def __init__(self, timeout, message=None, exception=TimeoutError,
                 caller=None):
        """
        :param timeout: Timeout period in seconds.
        :param message: Error message to include in assertion.
        :param exception: The class of exception to raise on timeout event.
        :param caller: The name of the calling function to log.
        """
        self.timeout = timeout
        self.message = message or MSG_TIMEOUT.format(
            caller or 'Method', timeout)
        self.exception = exception

    def handle_timeout(self, signum, frame):
        logger.info(self.message)
        raise self.exception(self.message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.timeout)

    def __exit__(self, type, value, traceback):
        signal.alarm(0)


def timeout(timeout=10, message=None, exception=TimeoutError):
    """Decorator to apply a timeout to a function.
    :param timeout: Number of seconds to wait.
    :param message: Error message to use.
    :param exception: The exception type to raise on event of timeout.
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            with TimeoutHandler(timeout, message, exception, func.__name__):
                return func(*args, **kwargs)
        return wraps(func)(wrapper)
    return decorator
