# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014, 2015, 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from autopilot.exceptions import StateNotFoundError
from retrying import retry
from autopilot import logging as autopilot_logging
from autopilot.input._common import get_center_point
from ubuntu_system_tests.helpers.unity8.exceptions import UnityException

from ubuntu_system_tests.helpers.autopilot import (
    patched_wait_select_single, validate_dbus_object)
from ubuntu_system_tests.helpers.ubuntuuitools.cpo.header import DashPageHeader  # NOQA
from ubuntu_system_tests.helpers.ubuntuuitools.cpo.common import CustomProxyObjectBase  # NOQA
from ubuntu_system_tests.helpers.unity8.utils import UNITY8_DASH_PATH_ROOT

SCOPE_CLASS_NAME = 'GenericScopeView'

logger = logging.getLogger(__name__)


class Dash(CustomProxyObjectBase):
    def __init__(self, *args):
        super().__init__(*args)
        self.dash_content_list = self.wait_select_single(
            'QQuickListView', objectName='dashContentList')

    def get_applications_grid(self):
        get_grid = self.get_scope('clickscope').wait_select_single(
            'CardGrid', objectName='local')
        return get_grid

    def get_application_icon(self, text):
        """Returns a 'Tile' icon that has the text 'text' from the application
        grid.

        :param text: String containing the text of the icon to search for.

        """
        app_grid = self.get_applications_grid()
        resp_grid = app_grid.wait_select_single('ResponsiveGridView')
        return resp_grid.select_single('Tile', text=text)

    def get_scope(self, scope_name='clickscope'):
        return self.dash_content_list.wait_select_single(
            'QQuickLoader', scopeId=scope_name)

    def get_scope_by_index(self, scope_index=0):
        return self.dash_content_list.wait_select_single(
            'QQuickLoader', objectName=("scopeLoader%i" % scope_index))

    @autopilot_logging.log_action(logger.info)
    def open_scope(self, scope_id):
        """Open a dash scope.

        :parameter scope_id: The id of the scope.
        :return: The scope.

        """
        scope_loader = self._get_scope_loader(scope_id)
        if scope_loader.isCurrent:
            logger.info('The scope is already open.')
            return self._get_scope_from_loader(scope_loader)
        else:
            return self._open_scope_scrolling(scope_loader)

    def _get_scope_loader(self, scope_id):
        try:
            aux = self.dash_content_list.get_children_by_type('QQuickItem')[0]
            for l in aux.get_children_by_type('QQuickLoader'):
                if (l.scopeId == scope_id):
                    return l
            raise UnityException(
                'No scope found with id {0}'.format(scope_id))
        except StateNotFoundError:
            raise UnityException(
                'No scope found with id {0}'.format(scope_id))

    def _get_scope_from_loader(self, loader):
        return loader.wait_select_single('GenericScopeView')

    def _open_scope_scrolling(self, scope_loader):
        scroll = self._get_scroll_direction(scope_loader)

        while not scope_loader.isCurrent:
            scroll()
            self.dash_content_list.moving.wait_for(False)

        scope_loader.isCurrent.wait_for(True)
        scope = self._get_scope_from_loader(scope_loader)
        return scope

    def _get_scroll_direction(self, scope_loader):
        current_scope_loader = self.dash_content_list.select_single(
            'QQuickLoader', isCurrent=True)
        if scope_loader.globalRect.x < current_scope_loader.globalRect.x:
            return self._scroll_to_left_scope
        elif scope_loader.globalRect.x > current_scope_loader.globalRect.x:
            return self._scroll_to_right_scope
        else:
            raise UnityException('The scope is already open')

    @retry(stop_max_attempt_number=3,
           wait_fixed=500,
           retry_on_exception=lambda exception: (
               isinstance(exception, AssertionError)))
    @autopilot_logging.log_action(logger.info)
    def _scroll_to_left_scope(self):
        original_index = self.dash_content_list.currentIndex
        dash_content = self.select_single(objectName="dashContent")
        x, y, width, height = dash_content.globalRect
        # Make the drag range be a multiple of the drag "rate" value.
        # Workarounds https://bugs.launchpad.net/mir/+bug/1399690
        rate = 5
        divisions = 5
        jump = (width / divisions) // rate * rate
        start_x = x + jump
        stop_x = x + jump * (divisions - 1)
        start_y = stop_y = y + 1
        self.pointing_device.drag(start_x, start_y, stop_x, stop_y, rate)
        self.dash_content_list.currentIndex.wait_for(original_index - 1)

    @retry(stop_max_attempt_number=3,
           wait_fixed=500,
           retry_on_exception=lambda exception: (
               isinstance(exception, AssertionError)))
    @autopilot_logging.log_action(logger.info)
    def _scroll_to_right_scope(self):
        original_index = self.dash_content_list.currentIndex
        dash_content = self.select_single(objectName="dashContent")
        x, y, width, height = dash_content.globalRect
        # Make the drag range be a multiple of the drag "rate" value.
        # Workarounds https://bugs.launchpad.net/mir/+bug/1399690
        rate = 5
        divisions = 5
        jump = (width / divisions) // rate * rate
        start_x = x + jump * (divisions - 1)
        stop_x = x + jump
        start_y = stop_y = y + 1
        self.pointing_device.drag(start_x, start_y, stop_x, stop_y, rate)
        self.dash_content_list.currentIndex.wait_for(original_index + 1)

    def enter_search_query(self, query, keyboard):
        current_header = self._get_current_page_header()
        search_button = \
            current_header.select_single(objectName="search_button")
        self.pointing_device.move(
            search_button.globalRect.x + search_button.width / 2,
            search_button.globalRect.y + search_button.height / 2)
        self.pointing_device.click()
        headerContainer = current_header.select_single(
            objectName="headerContainer")
        headerContainer.contentY.wait_for(0)
        keyboard.type(query)
        self.select_single(
            objectName="processingIndicator").visible.wait_for(False)

    def get_search_text_field(self):
        page_header = self._get_current_page_header()
        return page_header.select_single(objectName='searchTextField')

    def _get_current_page_header(self):
        all_scopes = self.dash_content_list.select_many("GenericScopeView")
        for i in all_scopes:
            if i.isCurrent:
                return i.select_single(objectName="scopePageHeader")
        return None

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, UNITY8_DASH_PATH_ROOT, b'Dash', objectName='dash')

    def wait_for_processing_to_complete(self, timeout=90):
        """Wait until the dash is fully loaded"""
        self.select_single(
            objectName='processingIndicator').visible.wait_for(False,
                                                               timeout=timeout)

    def wait_for_dash_loaded(self):
        """Wait until the dash content is ready"""
        dash_content = self.select_single(objectName='dashContent')
        dash_content.pageHeaderTotallyVisible.wait_for(True)

    def scroll_to_left_scope(self):
        self._scroll_to_left_scope()
        self.dash_content_list.moving.wait_for(False)

    def scroll_to_right_scope(self):
        self._scroll_to_right_scope()
        self.dash_content_list.moving.wait_for(False)

    def get_current_scope(self, cpo=None, *args, **kwargs):
        """
        Return the GenericScopeView proxy object for current scope.

        :param cpo: The required scope proxy object. If this is None the
                    returned CPO will be generic.
        :return: The required scope proxy object.

        """
        if not cpo:
            from ubuntu_system_tests.helpers.scopes.generic.cpo import (
                GenericScopeView)
            cpo = GenericScopeView
        return patched_wait_select_single(
            root=self, class_object=cpo, class_name=SCOPE_CLASS_NAME,
            isCurrent=True, *args, **kwargs)

    def get_scope_count(self):
        """Count how many scopes are present."""
        return len(self.dash_content_list.select_many(SCOPE_CLASS_NAME))

    def get_current_page_header(self, header_class=DashPageHeader,
                                wait_for_dash_loaded=True):
        """
        Return the PageHeader object representing the current header
        being displayed.

        :return: PageHeader object

        """
        if wait_for_dash_loaded:
            self.wait_for_dash_loaded()
        return self.get_current_scope().get_scope_page_header()

    def swipe_screen_from_left(self):
        """Move a focused application to the background and return to dash."""
        width = self.width
        height = self.height
        start_x = 0
        start_y = int(height/2)
        end_x = width
        end_y = start_y
        self.pointing_device.drag(start_x, start_y, end_x, end_y)

    def open_manage_scopes(self):
        """Open and return a helper for the scope management page."""
        from ubuntu_system_tests.helpers.scopes.scopes_list import ScopesList
        overview_drag_handle = self.wait_select_single(
            objectName='overviewDragHandle')
        start_x, start_y = get_center_point(overview_drag_handle)
        stop_x = start_x
        stop_y = 0
        self.pointing_device.drag(start_x, start_y, stop_x, stop_y)
        self.wait_for_dash_loaded()
        return self.wait_select_single(
            ScopesList,
            objectName='scopesList',
            visible=True)


class Preview(CustomProxyObjectBase):
    """Autopilot custom proxy object for generic previews."""
