#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014-2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import glob
import os

from collections import namedtuple
from softwareproperties.ppa import PPAException
from softwareproperties.shortcuts import ShortcutException
from softwareproperties.SoftwareProperties import shortcut_handler
from urllib.error import HTTPError

from ubuntu_system_tests.common import (
    config,
    ssh,
)
from ubuntu_system_tests.host import (
    debian,
    qemu,
)

# apt-cacher-ng definitions
APT_CACHER = 'apt-cacher-ng'
APT_CACHER_CACHE_DIR = '/var/cache/apt-cacher-ng'
APT_CACHER_LOG_DIR = '/var/log/apt-cacher-ng/'
APT_CACHER_UPSTART_INIT_FILE = '/etc/init/apt-cacher-ng.conf'
APT_CACHER_SYSTEMD_INIT_FILE = '/etc/systemd/system/apt-cacher-ng.service'
APT_PROXY = 'Acquire::http { Proxy \\"http://127.0.0.1:3142\\"; };'
APT_PROXY_FILE = '/etc/apt/apt.conf.d/01proxy'
EXT_CACHE_DIR = '/userdata/apt-cacher-ng/cache'
EXT_LOG_DIR = '/userdata/apt-cacher-ng/logs'
PERMISSION = '755'

QTTESTABILITY_AUTOPILOT = 'qttestability-autopilot'

DEFAULT_PPA_LIST = ['ppa:canonical-platform-qa/selenium',
                    'ppa:canonical-platform-qa/ubuntu-system-tests']
INSTALL_LIST = [QTTESTABILITY_AUTOPILOT]


_ADT_RUN_SHORT_TIMEOUT = 600  # 10 minutes.
_ADT_RUN_TEST_TIMEOUT = 86400  # 24 hours

REBOOT_FILE = '/tmp/ust-reboot-required'

STATUS_NO_CONNECTION = 1
STATUS_FS_RO_MOUNT_ERROR = 2
STATUS_FS_RW_MOUNT_ERROR = 3
STATUS_REBOOT_REQUIRED = 4
STATUS_PPA_SERIES_NOT_PUBLISHED = 5

TESTS_UST_DEFAULT = ['ubuntu_system_tests']
TESTS_SETUP = ['setup']
TESTS_SYSTEMTESTS = ['systemtests']

TARGET_ADB = 'adb'
TARGET_QEMU = 'qemu'
TARGETS = [TARGET_ADB, TARGET_QEMU]

Ppa = namedtuple('Ppa', ['id', 'dists_url'])
Ppa.__new__.__defaults__ = (None, None)

PPA_ID = 'ppa:'
PPA_ID_SILO = 'silo:'
PPA_CI_TRAIN_SILO = 'ppa:ci-train-ppa-service/{id}'


def build_basic_adt_run_test_command(config_stack, tests_to_run, verbose):
    """Build and return first part of adt-run command for test execution."""
    depends = debian.get_dependencies(tests_to_run or TESTS_UST_DEFAULT)
    print('using test dependencies: {}'.format(' '.join(depends)))
    run_cmd = _get_adt_run_cmd_start(config_stack, verbose)
    run_cmd.append('--override-control={}'.format(
        debian.create_temp_control_file(
            tests=TESTS_SYSTEMTESTS, dependencies=depends)))
    run_cmd.extend([
        # Updating and installing can take a while.
        '--timeout-short={}'.format(_ADT_RUN_SHORT_TIMEOUT),
        # Regression tests can take up to 8 hours, no reason to cut short.
        '--timeout-test={}'.format(_ADT_RUN_TEST_TIMEOUT),
        '-B',
        '--unbuilt-tree={}'.format(debian.get_debian_path()),
        # Even if the config file on the host has a different name, on the
        # testbed it's copied with the default name.
        '--copy',
        '{src}:{dest}'.format(
            src=config_stack.file_path,
            dest=config.get_device_config_file_path()),
    ])
    return run_cmd


def build_basic_adt_run_setup_command(config_stack, tests_to_run, verbose,
                                      cache_deps):
    """Build and return first part of adt-run command for device setup."""
    run_cmd = _get_adt_run_cmd_start(config_stack, verbose)
    if cache_deps:
        # Run the setup script from the control file listing all required
        # dependencies. This will ensure all required dependencies are
        # downloaded and cached as part of the setup command.
        depends = debian.get_dependencies(tests_to_run or TESTS_UST_DEFAULT)
        print('using test dependencies: {}'.format(' '.join(depends)))
        run_cmd.append('--override-control={}'.format(
            debian.create_temp_control_file(
                tests=TESTS_SETUP, dependencies=depends)))
    else:
        # Dependency caching is not required. In this case don't download
        # any dependencies by using a control file with no dependencies.
        run_cmd.append('--override-control={}'.format(
            debian.create_temp_control_file(tests=TESTS_SETUP)))

    run_cmd.extend([
        '-B',
        '--unbuilt-tree={}'.format(debian.get_debian_path())])
    return run_cmd


def build_adt_run_setup_command(config_stack, args):
    """Build and return adt-run command for device setup."""
    default_ppa_info = _validate_ppas(DEFAULT_PPA_LIST)
    ppas = _validate_ppas(_get_ppa_ids(args.ppa))
    run_cmd = build_basic_adt_run_setup_command(
        config_stack, args.suites, args.verbose, args.cache_deps)

    run_cmd.extend(_get_delete_reboot_required_setup_command())
    run_cmd.extend(_get_set_config_section_command(args.config_section))
    run_cmd.extend(
        _get_network_setup_command(config_stack, args.network_ready))
    run_cmd.extend(_get_disable_auto_reboot_setup_command())
    run_cmd.extend(_get_add_ppas_setup_command(default_ppa_info))
    run_cmd.extend(_get_add_ppas_setup_command(ppas))
    run_cmd.extend(_get_update_apt_setup_command(args.update_apt))
    run_cmd.extend(_get_install_test_pkgs_if_required_setup_command())
    run_cmd.extend(_get_apt_cacher_install_setup_command(args.cache_deps))
    run_cmd.extend(_get_dist_upgrade_if_required_setup_command(
        args.dist_upgrade, ppas))
    run_cmd.extend(
        _get_install_tests_setup_command(args.install_tests, args.suites))
    run_cmd.extend(_get_fs_ro_setup_command())
    run_cmd.extend(_get_apparmor_rules_update_setup_command())
    run_cmd.extend(_get_check_reboot_required_setup_command())

    target_cmd = _get_target_command(config_stack, args)

    return run_cmd + ['---'] + target_cmd


def build_adt_run_test_command(config_stack, args):
    """Build and return adt-run command for executing tests on device."""
    run_cmd = build_basic_adt_run_test_command(
        config_stack, args.suites, args.verbose)

    run_cmd.extend(_get_fs_ro_setup_command())
    run_cmd.extend(_get_set_tests_to_run_command(args.suites))
    run_cmd.extend(_get_set_config_section_command(args.config_section))
    run_cmd.extend(_get_set_touch_visualization(args.touch_visualization))
    run_cmd.extend(
        _get_network_setup_command(config_stack, args.network_ready))
    run_cmd.extend(_get_apparmor_rules_update_setup_command())
    run_cmd.extend(ssh.get_ssh_config_commands())

    target_cmd = _get_target_command(config_stack, args)

    return run_cmd + ['---'] + target_cmd


def _get_target_command(config_stack, args):
    """Return the autopkgtest virt-server command for specified target."""
    if args.target not in TARGETS:
        raise ValueError('{} is not a valid target type.'.format(args.target))
    if args.target == TARGET_ADB:
        return _get_ssh_command(config_stack)
    else:
        return _get_qemu_command(config_stack, args)


def _get_adt_run_cmd_start(config_stack, verbose):
    """Return the start of adt_run command."""
    output_dir = config_stack.get('output_dir')
    run_cmd = ['adt-run', '--output-dir', output_dir]
    if verbose:
        run_cmd.append('-ddd')
    return run_cmd


def _get_qemu_command(config_stack, args):
    """Return command to launch qemu with local image file."""
    img_path = args.img or os.path.join(
        config_stack.get('cache_data_path'), 'img')
    if os.path.isfile(img_path):
        # A file is specified so use that one
        img_file = img_path
    elif os.path.isdir(img_path):
        # Get the most recent img file from the directory
        try:
            img_file = max(
                glob.iglob(os.path.join(img_path, '*.img')),
                key=os.path.getctime)
        except ValueError:
            raise ValueError(
                'No image files found at {}.'.format(img_path))
    else:
        raise ValueError('{} does not exist.'.format(img_path))
    cpu = config_stack.get('qemu_cpu_count')
    ram = get_ram_kb(config_stack.get('qemu_ram_alloc'))
    headless = args.headless or config_stack.get(
        'qemu_headless').lower() == 'true'
    if headless:
        display = qemu.QEMU_DISPLAY_HEADLESS
    else:
        display = qemu.QEMU_DISPLAY_OPTS + qemu.QEMU_SOUND_OPTS
    return ([
        'qemu', img_file, '--cpus={}'.format(cpu), '--ram-size={}'.format(ram),
        '--qemu-options={}'.format(display)])


def get_ram_kb(ram):
    """Convert ram from G notation to number of KB, e.g. 4G = 4096."""
    if ram.endswith('G'):
        return str(int(ram[:-1]) * 1024)
    return ram


def _get_ppa_ids(csv_list):
    """Convert any ppa in csv_list from 'silo:' format to 'ppa:' url format.
    Alternatively if no 'ppa:' prefix is used, then apply this by default.
    :param csv_list: Command line ppa argument specifying string comma
                     separated value list of additional ppas in either 'ppa:',
                    'silo:' format or with no prefix.
    :return: List of converted ppa urls in 'ppa:' format.
    """
    formatted_list = []
    if not csv_list:
        return formatted_list
    for ppa in csv_list.split(','):
        formatted_list.append(_get_formatted_ppa(ppa))
    return formatted_list


def _get_formatted_ppa(ppa):
    """Return correctly formatted ppa id starting with 'ppa:' prefix. This will
    convert 'silo:' or no prefix to the standard 'ppa:' prefix.
    :param ppa: Unformatted ppa string.
    :return: ppa id starting with 'ppa:' prefix.
    """
    if ppa.startswith(PPA_ID_SILO):
        return PPA_CI_TRAIN_SILO.format(id=ppa.split(':')[1])
    elif not ppa.startswith(PPA_ID):
        return PPA_ID + ppa
    return ppa


def _get_ssh_command(config_stack):
    """Return command parameters for ssh connection to requried device."""
    device_password = config_stack.get('device_password')
    return (['ssh', '-s', 'adb', '--', '-p', device_password] +
            _get_serial_argument(config_stack))


def _get_adb_series_command(config_stack):
    """Return adb command to get series codename."""
    return _get_adb_command(
        config_stack, 'shell', 'lsb_release -cs | tr -d "\r"')


def _get_adb_reboot_commands(config_stack):
    """Return list of adb commands to reboot device."""
    return [
        _get_adb_command(config_stack, 'shell', 'sync'),
        _get_adb_command(config_stack, 'reboot')]


def _get_adb_start_server_command():
    """Return list of commands to start adb server on host."""
    return ['adb', 'start-server']


def _get_adb_command(config_stack, *args):
    """Return adb command to run specified command args on connected device."""
    return ['adb'] + _get_serial_argument(config_stack) + list(args)


def _get_serial_argument(config_stack):
    """Return serial commands from config, or empty list if not specified."""
    device_serial = config_stack.get('device_serial')
    if device_serial:
        return ['-s', device_serial]
    else:
        # No serial in the config, run the commands as if only one device is
        # connected.
        return []


def _get_set_tests_to_run_command(tests_to_run):
    """Return command to set list of tests as environment variable."""
    tests = ' '.join(tests_to_run or [])
    return _get_set_env_var_adt_command('TESTS_TO_RUN', tests)


def _get_set_config_section_command(config_section):
    """Return command to set config section name as environment variable."""
    return _get_set_env_var_adt_command('CONFIG_SECTION', config_section)


def _get_set_touch_visualization(touch_visualization):
    """Return command to set touch visualization as environment variable."""
    return _get_set_env_var_adt_command(
        'TOUCH_VISUALIZATION', str(touch_visualization).lower())


def _get_set_env_var_adt_command(name, value):
    """Return command to set environment variable on device."""
    return ['--env={n}={v}'.format(n=name, v=value)]


def _get_update_apt_command():
    """Return command to apt-get update."""
    return ('echo "apt-get updating"; apt-get update || '
            '(sleep 15; apt-get update); ')


def _get_update_apt_setup_command(update_apt):
    """Return setup command to apt-get update."""
    if update_apt:
        return [
            '--setup-commands',
            _get_fs_rw_command() +
            _get_update_apt_command()]
    return []


def _get_device_network_command(wifi_ssid, wifi_password):
    """Return command to configure wifi network on device.

    :param wifi_ssid: SSID of required wifi network
    :param wifi_password: Password of required wifi network

    """
    cmd_fmt = ('nmcli device wifi connect '
               '\'{ssid}\' password \'{password}\' 2>/dev/null || true')
    return cmd_fmt.format(ssid=wifi_ssid, password=wifi_password)


def _get_is_fs_ro_command():
    """Return command to indicate if file system is currently read-only."""
    return 'grep \' /\ ext\' /proc/mounts | grep -q \'ro\''


def _get_is_fs_rw_command():
    """Return command to indicate if file system is currently writable."""
    return 'grep \' /\ ext\' /proc/mounts | grep -q \'rw\''


def _get_fs_rw_command():
    """Return command to set file system read-write if not already."""
    return _get_if_command(_get_is_fs_ro_command(), _get_set_fs_rw_command())


def _get_fs_ro_setup_command():
    """Return setup command to set file system read-only if not already."""
    return [
        '--setup-commands',
        _get_if_command(_get_is_fs_rw_command(), _get_set_fs_ro_command())
        ]


def _get_set_fs_rw_command():
    """Return command to set file system writable, retyring on failure."""
    return ('echo "Mounting file system rw"; '
            'sync; RETRY=0; until mount -o remount,rw / || [ $RETRY -gt 10 ]; '
            'do sleep 1; RETRY=$(($RETRY+1)); echo "mount attempt failed, '
            'trying again..."; done; if [ $RETRY -gt 10 ]; then exit '
            '{status}; fi; '.format(status=STATUS_FS_RW_MOUNT_ERROR))


def _get_set_fs_ro_command():
    """Return command to set file system read-only, retyring on failure."""
    return ('echo "Mounting file system ro"; '
            'sync; RETRY=0; until mount -o remount,ro / || [ $RETRY -gt 10 ]; '
            'do sleep 1; RETRY=$(($RETRY+1)); echo "mount attempt failed, '
            'trying again..."; done; if [ $RETRY -gt 10 ]; then exit '
            '{status}; fi; '.format(status=STATUS_FS_RO_MOUNT_ERROR))


def _get_network_setup_command(config_stack, network_ready):
    """ The network needs to be configured on the device
    This setup-command should be the first to run so that wi-fi is ready
    """
    cmd = []
    if not network_ready:
        network_cmd = _get_device_network_command(
            config_stack.get('wifi_ssid'),
            config_stack.get('wifi_password'))
        cmd = ['--setup-commands', network_cmd]

    return cmd + [
        '--setup-commands', 'echo \'Checking device network connection...\'; '
        'RETRY=0; until wget -q --spider launchpad.net || [ $RETRY -gt 60 ]; '
        'do sleep 1; RETRY=$(($RETRY+1)); echo "wget attempt $RETRY failed, '
        ' trying again..."; done; if [ $RETRY -gt 60 ]; then echo \'Device '
        'has no network connection!\'; exit {status}; fi'.format(
            status=STATUS_NO_CONNECTION)
    ]


def _get_apt_cacher_autostart_command():
    return 'if [ -z "$(pidof systemd)" ]; then {} else {} fi;'.format(
        _get_apt_cacher_upstart_autostart_command(),
        _get_apt_cacher_systemd_autostart_command())


def _array_to_script(*lines):
    """ Retrieve a string with \n characters based on a string list
    :param lines: A list with the lines
    :return: the string with the array elements and \n characters
    """
    return '\n'.join(lines)


def _get_apt_cacher_upstart_autostart_command():
    """Write the apt-cacher-ng conf file into /etc/init/apt-cacher-ng.conf
    to re-launch apt-cacher-ng on each subsequent reboot.

    This method assumes that upstart is used as the init system. If this
    changes in future then this method will need updating to accomodate that.
    """
    script = _array_to_script(
        'start on runlevel [2345]',
        'script',
        _get_mount_dir_command(EXT_LOG_DIR, APT_CACHER_LOG_DIR) +
        _get_mount_dir_command(EXT_CACHE_DIR, APT_CACHER_CACHE_DIR) +
        _get_apt_cacher_start_command(),
        'end script')
    return _get_create_file_command(content=script,
                                    path=APT_CACHER_UPSTART_INIT_FILE)


def _get_apt_cacher_systemd_autostart_command():
    """Write the apt-cacher-ng conf file into /etc/init/apt-cacher-ng.conf
    to re-launch apt-cacher-ng on each subsequent reboot.

    This method assumes that upstart is used as the init system. If this
    changes in future then this method will need updating to accomodate that.
    """
    script = _array_to_script(
        '[Unit]',
        'Description=Autostart Apt Cacher NG',
        '',
        '[Service]',
        'Type=simple',
        'ExecStart={}',
        '',
        '[Install]',
        'WantedBy=multi-user.target').format(_get_apt_cacher_start_command())
    file_cmd = _get_create_file_command(content=script,
                                        path=APT_CACHER_SYSTEMD_INIT_FILE)
    enable_cmd = 'systemctl enable apt-cacher-ng.service;'
    return file_cmd + enable_cmd


def _get_apt_cacher_install_setup_command(apt_cacher_required):
    """Return command to install apt-cacher-ng on device and configure it to
    use userdata partion for caching and logging. If apt-cacher-ng is already
    installed on the device then no action will be taken."""
    if not apt_cacher_required:
        return []
    is_apt_cacher_installed = _get_packages_installed_command(APT_CACHER)
    install_apt_cacher_cmds = (
        _get_fs_rw_command() +
        _get_install_packages_command(APT_CACHER) +
        _get_apt_cacher_stop_command() +
        _get_create_folder_command(
            path=EXT_LOG_DIR, owner=APT_CACHER, permissions=PERMISSION) +
        _get_create_folder_command(
            path=EXT_CACHE_DIR, owner=APT_CACHER, permissions=PERMISSION) +
        _get_create_folder_command(
            path=APT_CACHER_CACHE_DIR, owner=APT_CACHER,
            permissions=PERMISSION) +
        _get_mount_dir_command(EXT_LOG_DIR, APT_CACHER_LOG_DIR) +
        _get_mount_dir_command(EXT_CACHE_DIR, APT_CACHER_CACHE_DIR) +
        _get_create_file_command(content=APT_PROXY, path=APT_PROXY_FILE) +
        _get_apt_cacher_start_command() +
        _get_apt_cacher_autostart_command())
    return ['--setup-commands', _get_if_not_command(
        is_apt_cacher_installed, install_apt_cacher_cmds)]


def _get_if_not_command(condition, commands):
    """Return if command for specified condition and commands."""
    return 'if ! ({condition}); then {commands} fi; '.format(
        condition=condition, commands=commands)


def _get_if_file_exists_command(file_path, commands, else_commands=None):
    """Return if file exists command to execute specified commands.
    :param file_path: Path of file to check if exists.
    :param commands: Commands to execute if file exists.
    :param else_commands: Optional commands to execute if file does not exist.
    :return: If command for specified file and actions.
    """
    command = 'if [ -f {p} ]; then {c} '.format(p=file_path, c=commands)
    if else_commands:
        command += 'else {c} '.format(c=else_commands)
    command += 'fi; '
    return command


def _get_if_command(condition, commands):
    """Return if not command for specified condition and commands."""
    return 'if ({condition}); then {commands} fi; '.format(
        condition=condition, commands=commands)


def _get_create_folder_command(path, owner, permissions):
    """Return command to create a folder with specified path, owner and
    permissions."""
    return ('if ! [ -d {path} ]; then '
            'mkdir -p {path}; fi; '
            'chown {owner}:{owner} {path}; '
            'chmod {permissions} {path}; ').format(
        path=path, owner=owner, permissions=permissions)


def _get_install_test_packages_command():
    """Return command to install required test packages."""
    return (_get_fs_rw_command() +
            _get_install_packages_command(' '.join(INSTALL_LIST)))


def _get_test_packages_installed_command():
    """Return command to indicate whether all required packages are already
    installed or not.
    :param apt_cacher_required: Whether apt-cahcer-ng is required to be
                                installed or not.
    """
    return _get_packages_installed_command(' '.join(INSTALL_LIST))


def _get_packages_installed_command(packages):
    """Return command to tell if all listed packages are installed or not."""
    return ('! apt-cache policy {packages} | '
            'grep -q \'Installed: (none)\'; '.format(packages=packages))


def _get_validate_ppa_series_command(ppa):
    """Return command to validate if ppa publishes for required series.
    :param ppa: Ppa tuple object for required ppa.
    :return: Command to validate ppa publishes for required series.
    """
    return ('wget -q --spider {u}$SERIES/ || ( echo "Error: {p} does not '
            'publish for series $SERIES."; exit {s}; ); '.format(
                u=ppa.dists_url, p=ppa.id, s=STATUS_PPA_SERIES_NOT_PUBLISHED))


def _get_add_ppa_required_command(ppa):
    """Return command to indicate if specified ppa is required to be added."""
    path = ppa.id.split(':')[1]
    return ('! grep -qrE "^deb\s.*{p}" /etc/apt/sources.list.d'.format(p=path))


def _get_add_ppa_command(ppa):
    """Return command to add specified ppa."""
    return 'apt-add-repository -y {p}; '.format(p=ppa.id)


def _get_add_ppas_command(ppa_list):
    """Return command to add each ppa in list if not already installed."""
    if not ppa_list:
        return ''
    cmd = _get_fs_rw_command() + _get_series_command()
    for ppa in ppa_list:
        cmd += _get_if_command(
            _get_add_ppa_required_command(ppa),
            _get_validate_ppa_series_command(ppa) +
            _get_add_ppa_command(ppa))
    return cmd


def _get_add_ppas_setup_command(ppas):
    """Return command to add each ppa in comma separated list."""
    if not ppas:
        return []
    return ['--setup-commands', _get_add_ppas_command(ppas)]


def _validate_ppas(ppa_ids):
    """Validate each ppa in list is valid and return list of Ppa objects."""
    ppa_list = []
    for ppa_id in ppa_ids:
        ppa_info = _get_ppa_info(ppa_id)
        if not ppa_info:
            raise NameError('{p} is not a valid ppa.'.format(p=ppa_id))
        ppa_dist_url = _get_ppa_dists_url(ppa_info)
        ppa_list.append(Ppa(ppa_id, ppa_dist_url))
    return ppa_list


def _get_ppa_info(ppa):
    """Return ppa info dictionary, or None if not found.
    :param ppa: required ppa id.
    """
    try:
        shortcut = shortcut_handler(ppa)
    except (HTTPError, PPAException, ShortcutException):
        return None
    else:
        return shortcut.info()


def _get_ppa_dists_url(ppa_info):
    """Return url of ppa which shows all available series."""
    private = 's' if ppa_info.get('private') else ''
    owner, dist, name = ppa_info['reference'].split('/')
    if owner.startswith('~'):
        owner = owner[1:]
    url = 'http{p}://ppa.launchpad.net/{o}/{n}/{d}/dists/'
    return url.format(p=private, o=owner, n=name, d=dist)


def _dist_upgrade_required_command():
    """Return command to indicate if any packages need to be updated as part
    of a dist-upgrade command."""
    return ('su - $ADT_NORMAL_USER -c \'apt-get dist-upgrade -s\' | '
            'grep -q \'Inst\\|Conf\\|Remv\'; ')


def _get_dist_upgrade_if_required_setup_command(required, ppas):
    """Return setup command to perform dist-upgrade if required.
    :param required: Whether dist-upgrade is specifically requested.
    :param ppas: List of PPAs to add.
    :return: Setup command to perform dist-upgrade.

    """
    if not required or not ppas:
        return []
    return ['--setup-commands', _get_dist_upgrade_command(ppas)]


def _get_install_test_pkgs_if_required_setup_command():
    """Return setup command to install test packages if required."""
    return ['--setup-commands',
            _get_if_not_command(
                _get_test_packages_installed_command(),
                _get_install_test_packages_command())]


def _get_install_packages_command(packages):
    """Return command to install packages on device."""
    return 'apt-get -y --no-install-recommends install {packages}; '.format(
        packages=packages)


def _get_mount_dir_command(target, source):
    """Return command to create a symlink between source and target
    directories."""
    return 'mount -o bind {target} {source}; '.format(
        target=target, source=source)


def _get_create_file_command(content, path):
    """Return command to create file with specified content and path."""
    return 'echo "{content}" | tee {path}; '.format(
        content=content, path=path)


def _get_apt_cacher_start_command():
    """Return command to start apt-cacher-ng process."""
    return ('/usr/sbin/apt-cacher-ng -c /etc/apt-cacher-ng '
            'pidfile=/var/run/apt-cacher-ng/pid '
            'SocketPath=/var/run/apt-cacher-ng/socket '
            'foreground=0; ')


def _get_apt_cacher_stop_command():
    """Return command to stop apt-cacher-ng process."""
    return 'pkill apt-cacher-ng || true; '


def _get_series_command():
    """Return command to save series of device."""
    return 'SERIES=$(lsb_release -cs | tr -d \'\\r\'); '


def _get_dist_upgrade_command(ppas):
    """Return command to dist-upgrade device.
    :param ppas: List of ppas that have been added to device.
    :return: Command to perform dist-upgrade.
    """
    cmd = ''
    if ppas:
        # Create apt preferences file to prioritise the PPA contents
        # and stop services re-starting before reboot using custom policy-rc.d
        cmd += (
            _get_series_command() +
            _get_if_file_exists_command(
                '/usr/sbin/policy-rc.d',
                'mv /usr/sbin/policy-rc.d /tmp/policy-rc.d; ') +
            _get_create_file_command(
                content='exit 101', path='/usr/sbin/policy-rc.d') +
            'chmod +x /usr/sbin/policy-rc.d; ' +
            _get_write_extra_ppas_pref_content_command(ppas) +
            'apt-get update; ')

    cmd += ('apt-mark hold lxc-android-config; ' +
            _get_if_command(
                _dist_upgrade_required_command(),
                _get_set_reboot_required_command() +
                _get_upgrade_command()) +
            'apt-mark unhold lxc-android-config; ')

    if ppas:
        # Restore previous policy-rc.d file or just remove the added one
        cmd += (
            _get_if_file_exists_command(
                '/tmp/policy-rc.d',
                'mv /tmp/policy-rc.d /usr/sbin/policy-rc.d; ',
                'rm /usr/sbin/policy-rc.d; '))
    return cmd


def _get_upgrade_command():
    """Return command to upgrade device."""
    return ('apt-get dist-upgrade --yes --force-yes; ' +
            'apt-get autoremove --yes --force-yes; ')


def _get_write_extra_ppas_pref_content_command(ppas):
    """Return command to create apt preferences file for upgrade.
    :param ppas: List of PPAs that have been added.
    :return: Command to create apt preferences file.
    """
    cmd = ''
    for ppa in ppas:
        path = ppa.id.split('/')[1]
        cmd += 'Package: *\nPin: release o=*{}*\nPin-Priority:1100\n\n'.format(
            path)
    cmd += 'Package: *\nPin: release a=$SERIES*\nPin-Priority: 50\n'
    return _get_create_file_command(
        content=cmd, path='/etc/apt/preferences.d/extra-ppas.pref')


def _get_check_reboot_required_setup_command():
    """If a reboot is required, then exit with status to indicate that."""
    return [
        '--setup-commands',
        'if [ -f {p} ]; then exit {s}; fi; '.format(
            p=REBOOT_FILE, s=STATUS_REBOOT_REQUIRED)]


def _get_set_reboot_required_command():
    """Create file to indicate that reboot is required."""
    return 'if [ ! -f {p} ]; then touch {p}; fi; '.format(p=REBOOT_FILE)


def _get_delete_reboot_required_setup_command():
    """Delete file to indicate that reboot is required."""
    return [
        '--setup-commands',
        'if [ -f {p} ]; then rm -f {p}; fi; '.format(p=REBOOT_FILE)]


def _get_disable_auto_reboot_setup_command():
    """Return setup command to disable automatic rebooting by autopkgtest."""
    return [
        '--setup-commands',
        'if [ ! -f /run/autopkgtest_no_reboot.stamp ]; then '
        'touch /run/autopkgtest_no_reboot.stamp; fi; ']


def _get_install_tests_setup_command(required, tests_to_run):
    """ Get a list with all the dependencies from the tests/control file and
    when the apt-get install does not install any package, the device is set
    as ro and exit, otherwise it is rebooted and will start as ro
    """
    if not required:
        return []
    dependencies = debian.get_dependencies(
        tests_to_run or TESTS_UST_DEFAULT,
        tmp_install=False, control_fmt=False)
    install_command = 'apt-get -o dir::cache::archives="/tmp/" -y install'
    clean_command = 'apt-get clean; '
    for dep in dependencies:
        install_command += ' {}'.format(dep.split()[0])
    install_command += '; sleep 5; '

    return ['--setup-commands',
            _get_fs_rw_command() +
            clean_command +
            install_command +
            clean_command]


def _get_apparmor_rules_update_setup_command():
    """Return setup command to update apparmor for autopilot introspection."""
    rules_file = '/var/cache/apparmor/click-ap.rules'
    update_command = 'printf "dbus (receive, send) ' \
                     'bus=session path=/com/canonical/Autopilot/**,\n' \
                     '/tmp/adt-run.** r,\n" > {rules};' \
                     'aa-clickhook --include={rules}'.format(rules=rules_file)
    return [
        '--setup-commands',
        'if [ ! -e {rules} ];then {command};fi'.format(rules=rules_file,
                                                       command=update_command)
    ]
