#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014-2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import glob
import json
import os
import tempfile

from collections import defaultdict

# Installation path for shared files
INSTALLED_SHARE_DIR = '/usr/share/ubuntu-system-tests'
# Installation path for installed python scripts
INSTALLED_PYTHON_DIR = '/usr/lib/python'


def get_dependencies(tests_to_run, tmp_install=True, control_fmt=True):
    """
    Return the list of dependencies required for specified tests.

    :param tests_to_run: List of test ids to run.
    :param tmp_install: Whether to include dependencies which should only
                        be installed in /tmp location on device.
    :param control_fmt: Whether the returned list of dependencies should be in
                        control file format or command line format. This
                        affects the delimeter used and whether the version
                        is specified or not. True indicates control file
                        format, False indicates command line format.
    :return: Sorted list of required dependencies.
    :raises: RuntimeError if no matching dependencies are found.

    """
    depends_map = _read_dependency_files(
        os.path.realpath('debian/tests'), tmp_install, control_fmt)
    depends = []
    for test in tests_to_run:
        depends.extend(_dependencies_for_test(test, depends_map))
    return sorted(list(set(depends)))


def _read_dependency_files(deps_path, tmp_install, control_fmt):
    """
    Read the depenency data from .json files in the specified directory.

    :param deps_path: The path to the directory containing json files.
    :param tmp_install: Whether to include tmpinstall dependencies.
    :param control_fmt: Whether deps are in control file or command line
                        format.
    :returns: A dictionary associating test prefixes with lists of
              dependencies.
    """
    depends_map = defaultdict(list)
    for file in glob.glob(deps_path + '/*.json'):
        with open(file) as deps_file:
            jsn = json.loads(deps_file.read())
            _assert_no_duplicate_dep_id(jsn)
            deps_common = jsn.get('dependencies-common', [])
            for test in jsn['tests']:
                dep_ids = sorted(deps_common + test['depends'])
                deps = _get_formatted_dependency_list(
                    dep_ids, jsn, tmp_install, control_fmt)
                for id in test['ids']:
                    depends_map[id].extend(deps)
    return depends_map


def _dependencies_for_test(test_id, depends_map):
    """
    Return a list of dependencies required for the specified test id.

    :param test_id: The test id to get the dependencies for.
    :param depends_map: The dependency mapping dictionary.
    :returns: A list of required dependencies.

    """
    dependencies = []
    for id in depends_map.keys():
        if test_id.startswith(id) or id.startswith(test_id):
            # The start of test_id matches with a longer defined id
            # or a defined id matches the start of a longer test_id.
            # Either way, the dependencies should be added to list.
            dependencies.extend(depends_map[id])
    if not dependencies:
        raise RuntimeError(
            'No dependencies defind for test: {t}. Check '
            'debian/tests/dependencies.json.'.format(t=test_id))
    return dependencies


def _get_formatted_dependency_list(deps, jsn, tmp_install, control_fmt):
    """Return a list of formatted dependencies from specified list.

    :param deps: List of required dependency ids.
    :param jsn: Entire dictionary of dependencies loaded from json document.
    :param tmp_install: Whether to include dependencies that are only to
                        be installed in /tmp location.
    :param control_fmt: Whether the format of the dependencies should be in
                        control file format or command line format. True
                        indicates control file format, False is command line.
    :return: List of formatted dependencies, including version if required.

    """
    dep_list = []
    for dep in deps:
        dep_list.extend(
            _get_formatted_dependency(dep, jsn, tmp_install, control_fmt))
    return dep_list


def _assert_no_duplicate_dep_id(jsn):
    """
    Assert there are no duplicate dependency definitions in json dictionary.

    :param jsn: json dictionary containing dependency definitions.
    :raises: ValueError if more than one definition is present for a single
             dependency.

    """
    dep_ids = {}
    for dep in jsn['dependencies']:
        dep_id = dep['id']
        if dep_id in dep_ids:
            raise ValueError('Multiple dependencies defined for: '
                             '{d}. Please correct definition in json '
                             'document.'.format(d=dep_id))
        else:
            dep_ids[dep_id] = True


def _get_formatted_dependency(dependency_id, jsn, tmp_install, control_fmt):
    """Return the formatted dependecy string for provided id.

    :param dependency_id: Required dependency id.
    :param jsn: Entire dictionary of dependencies loaded from json document.
    :param tmp_install: Whether to include dependencies that are only to
                        be installed in /tmp location.
    :param control_fmt: Whether the format of the dependencies should be in
                        control file format or command line format. True
                        indicates control file format, False is command line.
    :return: List of formatted dependencies, including version if required. If
             dependency is not required then return empty list.

    """
    for dep in jsn['dependencies']:
        if dep['id'] == dependency_id:
            if is_dependency_required(dep, tmp_install):
                return [_get_dependency_str(dep, control_fmt)]
    return []


def is_dependency_required(dep, tmp_install):
    """Determine if the specified dependency is required or not.

    :param dep: Dependency dictionary definition.
    :param tmp_install: Whether to include dependencies that are only to
                        be installed in /tmp location.
    :return: True if required, False if not.

    """
    if not tmp_install and dep.get('tmpinstall', 'false') == 'true':
        # This is a tmp_install item and it is not required
        return False
    if dep.get('condition') == 'installed':
        # This is a special case where the dependency is only required when
        # running from installed location
        return _is_installed()
    # No other conditions to check, so this dependency must be required
    return True


def _get_dependency_str(dep_jsn, control_fmt):
    """Return the formatted dependency string for provided dependency,
    including version if required.

    :param dep_jsn: Dictionary containting dependency definition.
    :param control_fmt: Whether the string should be formatted in control file
                        format or command line format.
    :return: Formatted dependency string.

    """
    dep_str = dep_jsn['id']
    if dep_jsn.get('version') and control_fmt:
        # Add the control file format version specifier.
        # Note that the version specifier is not used for the other case
        # of apt-get command line install, as this doesn't support
        # minimum version specifier, only absolute version.
        dep_str += ' (>={})'.format(dep_jsn['version'])
    # Finally add delimeter
    dep_str += ',' if control_fmt else ' '
    return dep_str


def create_temp_control_file(tests=None, restrictions=None, dependencies=None):
    """Create a temp control file, appending the list of
    dependencies

    :param dependencies: List of dependencies to append to file
    :returns: File path to the new control file
    """
    if not dependencies:
        dependencies = []
    content = 'Tests: ' + ', '.join(tests or ['systemtests', 'setup'])
    content += '\nRestrictions: ' + ', '.join(restrictions or ['allow-stderr'])
    content += '\nDepends: ' + '\n         '.join(dependencies or [])
    content += '\n'
    return create_temp_file(content)


def create_temp_file(content):
    """Create a temporary file with specified content and return path to it."""
    _, tmp_path = tempfile.mkstemp(text=True)
    with open(tmp_path, 'w') as tmp_file:
        tmp_file.write(content)
    return tmp_path


def _is_installed():
    """Return True if the suite is running from installed location."""
    return os.path.realpath(__file__).startswith(INSTALLED_PYTHON_DIR)


def get_debian_path():
    """Return the path of the debian folder.
    This will depend on whether tests are running from installed location
    or local source directory.
    """
    return INSTALLED_SHARE_DIR if _is_installed() else os.getcwd()
