# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015-2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
from unittest.mock import (
    Mock,
    patch,
)

import testscenarios

from subprocess import CalledProcessError

from ubuntu_system_tests.host import command_line
from ubuntu_system_tests.host.command_line import (
    FIELD_EXTERNAL_ID,
    FIELD_VALUE,
)
from ubuntu_system_tests.common import config
from ubuntu_system_tests.selftests.utils import (
    get_cmdline_list_params,
    get_cmdline_run_params,
    get_cmdline_setup_params,
    ConfigBaseTestCase)

MULTI_TESTS = [
    {FIELD_EXTERNAL_ID: {FIELD_VALUE: 'module1.file1.class1.test1'}},
    {FIELD_EXTERNAL_ID: {FIELD_VALUE: 'module2.file1.class1.test1'}},
    {FIELD_EXTERNAL_ID: {FIELD_VALUE: 'module3.file1.class1.test1'}},
]

SINGLE_TEST = {
    FIELD_EXTERNAL_ID: {FIELD_VALUE: 'module1.file1.class1.test1'}
}


MULTI_TESTS_MISSING_EXT_ID = [
    {FIELD_EXTERNAL_ID: {FIELD_VALUE: 'module1.file1.class1.test1'}},
    {'dummy_field': {FIELD_VALUE: 'module2.file1.class1.test1'}},
    {FIELD_EXTERNAL_ID: {FIELD_VALUE: 'module3.file1.class1.test1'}},
]

MULTI_TESTS_MISSING_VALUE = [
    {FIELD_EXTERNAL_ID: {FIELD_VALUE: 'module1.file1.class1.test1'}},
    {FIELD_EXTERNAL_ID: {'dummy_field': 'module2.file1.class1.test1'}},
    {FIELD_EXTERNAL_ID: {FIELD_VALUE: 'module3.file1.class1.test1'}},
]

MULTI_TESTS_RESULT = [
    'module1.file1.class1.test1',
    'module2.file1.class1.test1',
    'module3.file1.class1.test1'
]

MULTI_TESTS_INVALID_FIELD_RESULT = [
    'module1.file1.class1.test1',
    'module3.file1.class1.test1'
]

SINGLE_TEST_RESULT = [
    'module1.file1.class1.test1'
]

DUMMY_PPA_INFO = {'private': False, 'reference': '~user/dist/name'}


class SetupSystemTestsCommandsTestCase(ConfigBaseTestCase,
                                       testscenarios.TestWithScenarios):

    scenarios = [
        # The nmcli network setup command is passed as --setup-command option
        # to adt-run. Verify that it should not run as separate command.
        ('without network setup', {
            'keyword_arguments': {'target': 'adb',
                                  'update_apt': False,
                                  'network_ready': True},
            'expected_commands': ['adt-run'],
        }),
        ('with network setup', {
            'keyword_arguments': {'target': 'adb',
                                  'update_apt': False,
                                  'network_ready': False},
            'expected_commands': ['adt-run'],
        }),
    ]

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()

    @patch('ubuntu_system_tests.host.command_line.generate_user_report')
    @patch('ubuntu_system_tests.host.commands._get_ppa_info',
           return_value=DUMMY_PPA_INFO)
    @patch('ubuntu_system_tests.host.debian.get_dependencies',
           return_value=['dep1'])
    def test_setup_system_tests_commands(
            self, mock_get_dependencies, mock_get_ppa_info,
            mock_generate_user_report):
        def get_command_from_call(call):
            """Return the name of the command, without the arguments."""
            cmd_list = call[0][0]
            if type(cmd_list) == str:
                cmd_list = cmd_list.split(' ')
            return cmd_list[0]

        self.set_config_values()
        conf_name = os.path.basename(self.config_stack.file_path)
        params = get_cmdline_setup_params(
            config=conf_name, **self.keyword_arguments)

        # Do not execute the subprocesses, just record their calls.
        with patch('subprocess.Popen') as mock_popen:
            enter = mock_popen().__enter__()
            # Make sure that check_output and call return with no errors.
            enter.communicate = Mock(return_value=(0, 'dummy'))
            enter.poll = Mock(return_value=0)
            enter.wait = Mock(return_value=0)
            command_line.setup_system_tests(params)

        # The first call doesn't run any command.
        calls = mock_popen.call_args_list[1:]
        commands = [get_command_from_call(call) for call in calls]
        self.assertEqual(self.expected_commands, commands)


class ListCommandLineArgumentsTestCase(testscenarios.TestWithScenarios):

    LIST_CMD = 'list'
    DEFAULT_CONFIG_FILE = 'ubuntu-system-tests.conf'
    DEFAULT_CONFIG_SECT = config.KEY_DEFAULT
    ALL_SUITES = []
    DO_USE_SILENT_MODE = True
    DONT_USE_SILENT_MODE = False

    scenarios = [
        ('no arguments', {
            'command_line_arguments': [],
            'parsed_arguments': (
                LIST_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                DONT_USE_SILENT_MODE,
                ALL_SUITES,
            )
        }),
        ('silent mode', {
            'command_line_arguments': ['--silent'],
            'parsed_arguments': (
                LIST_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                DO_USE_SILENT_MODE,
                ALL_SUITES,
            )
        }),
        ('single suite parameter', {
            'command_line_arguments': ['suite.tests.test_1'],
            'parsed_arguments': (
                LIST_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                DONT_USE_SILENT_MODE,
                ['suite.tests.test_1'],
            )
        }),
        ('multiple suite parameter', {
            'command_line_arguments': ['suite1.tests', 'suite2.tests'],
            'parsed_arguments': (
                LIST_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                DONT_USE_SILENT_MODE,
                ['suite1.tests', 'suite2.tests'],
            )
        }),
        ('string suite parameter', {
            'command_line_arguments': ['suite1.tests suite2.tests'],
            'parsed_arguments': (
                LIST_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                DONT_USE_SILENT_MODE,
                ['suite1.tests', 'suite2.tests'],
            )
        }),
        ('filter', {
            'command_line_arguments': ['--filters=filter:12345'],
            'parsed_arguments': (
                LIST_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                DONT_USE_SILENT_MODE,
                ALL_SUITES,
                'filter:12345',
            )
        }),
        ('filter and suites', {
            'command_line_arguments': [
                '--filters=filter:12345', 'suite1.tests', 'suite2.tests'],
            'parsed_arguments': (
                LIST_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                DONT_USE_SILENT_MODE,
                ['suite1.tests', 'suite2.tests'],
                'filter:12345',
            )
        }),
    ]

    @patch('ubuntu_system_tests.host.command_line.list_system_tests')
    def test_main_must_pass_command_line_arguments_to_list_system_tests(
            self, mock_list_system_tests):
        command_line.main([self.LIST_CMD] + self.command_line_arguments)
        params = get_cmdline_list_params(*self.parsed_arguments)
        mock_list_system_tests.assert_called_once_with(params)


class SetupCommandLineArgumentsTestCase(testscenarios.TestWithScenarios):

    SETUP_CMD = 'setup'
    DEFAULT_CONFIG_FILE = 'ubuntu-system-tests.conf'
    DEFAULT_CONFIG_SECT = config.KEY_DEFAULT
    ALL_SUITES = []
    DO_APT_UPDATE = True
    DONT_APT_UPDATE = False
    NETWORK_ALREADY_SETUP = True
    NETWORK_NOT_ALREADY_SETUP = False
    DO_DIST_UPGRADE = True
    DONT_DIST_UPGRADE = False
    DO_INSTALL_TESTS = True
    DONT_INSTALL_TESTS = False
    DO_USE_CACHE_DEPS = True
    DONT_USE_CACHE_DEPS = False
    DO_USE_SILENT_MODE = True
    DONT_USE_SILENT_MODE = False
    DO_USE_VERBOSE_LOGGING = True
    DONT_USE_VERBOSE_LOGGING = False
    NO_FILTER = None
    DEFAULT_TARGET = 'adb'
    DO_USE_HEADLESS = True
    DONT_USE_HEADLESS = False
    DEFAULT_IMG = None

    scenarios = [
        ('default setup arguments', {
            'command_line_arguments': [],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
            )
        }),
        ('alternative config', {
            'command_line_arguments': ['--config=alternative.conf'],
            'parsed_arguments': (
                SETUP_CMD,
                'alternative.conf',
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
            )
        }),
        ('alternative config section', {
            'command_line_arguments': ['--config-section=custom-section'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                'custom-section',
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
            )
        }),
        ('only apt-get update', {
            'command_line_arguments': ['--update-apt'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_CACHE_DEPS,
                DONT_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
            )
        }),
        ('with network already setup', {
            'command_line_arguments': ['--network-ready'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
            )
        }),
        ('with dist-upgrade enabled', {
            'command_line_arguments': ['--dist-upgrade'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
            )
        }),
        ('with install dependencies enabled', {
            'command_line_arguments': ['--install-tests'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_CACHE_DEPS,
                DONT_DIST_UPGRADE,
                DO_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
            )
        }),
        ('with cache deps specified', {
            'command_line_arguments': ['--cache-deps'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DONT_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
            )
        }),
        ('with verbose logging specified', {
            'command_line_arguments': ['--verbose'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DO_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
            )
        }),
        ('single test', {
            'command_line_arguments': ['suite.tests.test_1'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ['suite.tests.test_1'],
            )
        }),
        ('multiple tests', {
            'command_line_arguments': [
                'suite.tests.test_1', 'suite.tests.test_2'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ['suite.tests.test_1', 'suite.tests.test_2'],
            )
        }),
        ('string formatted tests', {
            'command_line_arguments': [
                'suite.tests.test_1 suite.tests.test_2'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ['suite.tests.test_1', 'suite.tests.test_2'],
            )
        }),
        ('filters', {
            'command_line_arguments': ['--filters=filter:12345'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
                'filter:12345',
            )
        }),
        ('extra ppa', {
            'command_line_arguments': ['--ppa=ppa:user/myppa'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
                NO_FILTER,
                'ppa:user/myppa',
            )
        }),
    ]

    @patch('ubuntu_system_tests.host.command_line.run_autopkg_setup')
    def test_main_must_pass_command_line_arguments_to_run_system_tests(
            self, mock_run_autopkg_setup):
        command_line.main([self.SETUP_CMD] + self.command_line_arguments)
        params = get_cmdline_setup_params(*self.parsed_arguments)
        mock_run_autopkg_setup.assert_called_once_with(params)


@patch('ubuntu_system_tests.host.command_line.print_warnings')
@patch('ubuntu_system_tests.host.command_line._prepare_config')
@patch('ubuntu_system_tests.host.command_line.add_tests_ids_for_filters')
class RunSystemTests(ConfigBaseTestCase):

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()

    def test_no_tests_for_filter_or_suites_raises(
            self, mock_add_tests_ids_for_filters, mock_prepare_config,
            mock_print_warnings):
        mock_prepare_config.return_value = self.config_stack
        mock_add_tests_ids_for_filters.return_value = False
        args = get_cmdline_run_params()
        with self.assertRaises(RuntimeError):
            command_line.run_system_tests(args)


class RunCommandLineArgumentsTestCase(testscenarios.TestWithScenarios):

    RUN_CMD = 'run'
    DEFAULT_CONFIG_FILE = 'ubuntu-system-tests.conf'
    DEFAULT_CONFIG_SECT = config.KEY_DEFAULT
    ALL_SUITES = []
    NETWORK_ALREADY_SETUP = True
    NETWORK_NOT_ALREADY_SETUP = False
    DO_USE_VERBOSE_LOGGING = True
    DONT_USE_VERBOSE_LOGGING = False
    DO_USE_TOUCH_VISUALIZATION = True
    DONT_USE_TOUCH_VISUALIZATION = False
    DO_USE_SILENT_MODE = True
    DONT_USE_SILENT_MODE = False
    DEFAULT_TARGET = 'adb'
    DO_USE_HEADLESS = True
    DONT_USE_HEADLESS = False
    DEFAULT_IMG = None

    scenarios = [
        ('default run arguments', {
            'command_line_arguments': [],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('alternative config', {
            'command_line_arguments': ['--config=alternative.conf'],
            'parsed_arguments': (
                RUN_CMD,
                'alternative.conf',
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('alternative config section', {
            'command_line_arguments': ['--config-section=custom-section'],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                'custom-section',
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('with network already setup', {
            'command_line_arguments': ['--network-ready'],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('with verbose logging specified', {
            'command_line_arguments': ['--verbose'],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DO_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('single test', {
            'command_line_arguments': ['suite.tests.test_1'],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ['suite.tests.test_1'],
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('mutliple tests', {
            'command_line_arguments': [
                'suite.tests.test_1', 'suite.tests.test_2'],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ['suite.tests.test_1', 'suite.tests.test_2'],
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('string formatted tests', {
            'command_line_arguments': [
                'suite.tests.test_1 suite.tests.test_2'],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ['suite.tests.test_1', 'suite.tests.test_2'],
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('filter', {
            'command_line_arguments': [
                '--filters=filter:12345'],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                DEFAULT_TARGET,
                DONT_USE_HEADLESS,
                DEFAULT_IMG,
                ALL_SUITES,
                DONT_USE_TOUCH_VISUALIZATION,
                'filter:12345',
            )
        }),
    ]

    @patch('ubuntu_system_tests.host.command_line.run_system_tests')
    def test_main_must_pass_command_line_arguments_to_run_system_tests(
            self, mock_run_system_tests):
        command_line.main([self.RUN_CMD] + self.command_line_arguments)
        params = get_cmdline_run_params(*self.parsed_arguments)
        mock_run_system_tests.assert_called_once_with(params)


class ValidFilterTypeValidationTestCase(testscenarios.TestWithScenarios):

    scenarios = [
        ('test case', {
            'filter_str': 'test:1234',
            'expected_result': ('test', '1234')
        }),
        ('filter', {
            'filter_str': 'filter:1234',
            'expected_result': ('filter', '1234')
        }),
        ('default type', {
            'filter_str': '1234',
            'expected_result': ('filter', '1234')
        }),
        ('sanity alias', {
            'filter_str': 'sanity',
            'expected_result': ('filter', '63805')
        }),
        ('regression alias', {
            'filter_str': 'regression',
            'expected_result': ('filter', '117776')
        }),
    ]

    def test_filter_validation(self):
        result = command_line._validate_filter(self.filter_str)
        self.assertEqual(result, self.expected_result)


class InvalidFilterTypeValidationTestCase(testscenarios.TestWithScenarios):

    scenarios = [
        ('test case not number', {
            'filter_str': 'test:1234ABC',
        }),
        ('filter not number', {
            'filter_str': 'filter:1234ABC',
        }),
        ('invalid type', {
            'filter_str': 'XYZ:1234',
        }),
        ('no type id not number', {
            'filter_str': '1234ABC',
        }),
    ]

    def test_invalid_filter_raises(self):
        with self.assertRaises(ValueError):
            command_line._validate_filter(self.filter_str)


@patch('ubuntu_system_tests.host.command_line._get_test_ids_for_filters')
class AddTestsFromFiltersTestCase(ConfigBaseTestCase,
                                  testscenarios.TestWithScenarios):

    scenarios = [
        ('tests added', {
            'test_ids': ['suite1.tests.test_1', 'suite2.tests.test_1'],
            'filters': '123',
            'result': True
        }),
        ('no tests found', {
            'test_ids': [],
            'filters': '123',
            'result': False
        }),
        ('no filter', {
            'test_ids': [],
            'filters': None,
            'result': True
        }),
    ]

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()
        self.set_config_values()

    def test_add_test_ids_for_filters(self, mock_get_test_ids):
        mock_get_test_ids.return_value = self.test_ids
        params = get_cmdline_run_params(filters=self.filters)
        result = command_line.add_tests_ids_for_filters(
            self.config_stack, params)
        self.assertEqual(result, self.result)
        self.assertEqual(params.suites, self.test_ids)


@patch('ubuntu_system_tests.host.practitest.PractitestSession.get_tests')
@patch('ubuntu_system_tests.host.practitest.PractitestSession.get_test')
class GetTestIDsForFiltersTestCase(ConfigBaseTestCase,
                                   testscenarios.TestWithScenarios):

    scenarios = [
        ('single test filter', {
            'get_test_result': SINGLE_TEST,
            'get_tests_result': None,
            'filters': 'test:123',
            'result': SINGLE_TEST_RESULT
        }),
        ('multiple test filters', {
            'get_test_result': SINGLE_TEST,
            'get_tests_result': None,
            'filters': 'test:123,test:456',
            'result': SINGLE_TEST_RESULT + SINGLE_TEST_RESULT
        }),
        ('single filter', {
            'get_test_result': None,
            'get_tests_result': MULTI_TESTS,
            'filters': 'filter:123',
            'result': MULTI_TESTS_RESULT
        }),
        ('multiple filters', {
            'get_test_result': None,
            'get_tests_result': MULTI_TESTS,
            'filters': 'filter:123,filter:456',
            'result': MULTI_TESTS_RESULT + MULTI_TESTS_RESULT
        }),
        ('test and filter', {
            'get_test_result': SINGLE_TEST,
            'get_tests_result': MULTI_TESTS,
            'filters': 'test:123,filter:456',
            'result': SINGLE_TEST_RESULT + MULTI_TESTS_RESULT
        }),

    ]

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()
        self.set_config_values()

    def test_add_test_ids_for_filters(self, mock_get_test, mock_get_tests):
        mock_get_test.return_value = self.get_test_result
        mock_get_tests.return_value = self.get_tests_result
        result = command_line._get_test_ids_for_filters(
            self.config_stack, self.filters)
        self.assertEqual(result, self.result)


class GetIDsForTestTestCase(testscenarios.TestWithScenarios):

    scenarios = [
        ('multiple tests', {
            'tests': MULTI_TESTS,
            'result': MULTI_TESTS_RESULT
        }),
        ('single test', {
            'tests': [SINGLE_TEST],
            'result': SINGLE_TEST_RESULT
        }),
        ('test missing external id', {
            'tests': MULTI_TESTS_MISSING_EXT_ID,
            'result': MULTI_TESTS_INVALID_FIELD_RESULT
        }),
        ('test missing value', {
            'tests': MULTI_TESTS_MISSING_VALUE,
            'result': MULTI_TESTS_INVALID_FIELD_RESULT
        }),
    ]

    def test_get_test_ids_for_tests(self):
        result = command_line._get_ids_for_tests(self.tests)
        self.assertEqual(result, self.result)


@patch('ubuntu_system_tests.host.command_line._prepare_config')
@patch('ubuntu_system_tests.host.command_line._get_test_ids_for_filters')
@patch('ubuntu_system_tests.host.command_line._get_test_ids_for_suites')
@patch('builtins.print')
class ListSystemTestsTestCase(ConfigBaseTestCase,
                              testscenarios.TestWithScenarios):

    scenarios = [
        ('filter only', {
            'tests_from_filter': ['filter_test_1', 'filter_test_2'],
            'tests_from_suite': [],
            'filters': 'filter:1234',
            'suites': '',
            'result': ['filter_test_1', 'filter_test_2'],
        }),
        ('suite only', {
            'tests_from_filter': [],
            'tests_from_suite': ['suite.test_1', 'suite.test_2'],
            'filters': '',
            'suites': 'suite1',
            'result': ['suite.test_1', 'suite.test_2'],
        }),
        ('filter and suite', {
            'tests_from_filter': ['filter_test_1', 'filter_test_2'],
            'tests_from_suite': ['suite1.test_1', 'suite1.test_2'],
            'filters': 'filter:1234',
            'suites': 'suite1',
            'result': ['filter_test_1', 'filter_test_2',
                       'suite1.test_1', 'suite1.test_2'],
        }),
        ('no duplicate ids', {
            'tests_from_filter': ['filter_test_1', 'test_2'],
            'tests_from_suite': ['suite1.test_1', 'test_2'],
            'filters': 'filter:1234',
            'suites': 'suite1',
            'result': ['filter_test_1', 'suite1.test_1', 'test_2'],
        }),
        ('ordered results', {
            'tests_from_filter': ['filter_test_2', 'filter_test_1'],
            'tests_from_suite': ['suite1.test_2', 'suite1.test_1'],
            'filters': 'filter:1234',
            'suites': 'suite1',
            'result': ['filter_test_1', 'filter_test_2',
                       'suite1.test_1', 'suite1.test_2'],
        }),

    ]

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()
        self.set_config_values()

    def test_list_tests(
            self, mock_print, mock_get_test_ids_for_suites,
            mock_get_test_ids_for_filters, mock_prepare_config):
        mock_prepare_config.return_value = self.config_stack
        mock_get_test_ids_for_suites.return_value = self.tests_from_suite
        mock_get_test_ids_for_filters.return_value = self.tests_from_filter
        args = get_cmdline_list_params(
            filters=self.filters, suites=self.suites)
        command_line.list_system_tests(args)
        mock_print.assert_called_once_with('\n'.join(self.result))


class TestListAutopilotTestCase(ConfigBaseTestCase,
                                testscenarios.TestWithScenarios):

    scenarios = [
        ('single test', {
            'output': b'suite1.test1\n',
            'tests': ['suite1.test1']
        }),
        ('multiple tests', {
            'output': b'suite1.test1\nsuite1.test2\n',
            'tests': ['suite1.test1', 'suite1.test2']
        }),
        ('blank lines', {
            'output': b'\n\nsuite1.test1\n\nsuite1.test2\n\n',
            'tests': ['suite1.test1', 'suite1.test2']
        }),
        ('whitespace', {
            'output': b' suite1.test1\n \nsuite1.test2 \n \n',
            'tests': ['suite1.test1', 'suite1.test2']
        }),
        ('chars and whitespace', {
            'output': b' loading tests from \nsuite1.test1\ntotal tests\n',
            'tests': ['suite1.test1']
        }),
    ]

    def test_get_ids_from_autopilot_output(self):
        with patch('subprocess.check_output', return_value=self.output):
            tests = command_line._list_autopilot_tests('suite')
            self.assertEqual(tests, self.tests)


class TestListAutopilotSuitesTestCase(ConfigBaseTestCase,
                                      testscenarios.TestWithScenarios):

    scenarios = [
        ('single suite', {
            'suites': ['suite1'],
            'output': [b'suite1.test1'],
            'tests': ['suite1.test1'],
        }),
        ('multiple suites', {
            'suites': ['suite1', 'suite2'],
            'output': [b'suite1.test1', b'suite2.test1'],
            'tests': ['suite1.test1', 'suite2.test1'],
        }),
    ]

    def test_get_ids_from_autopilot_output(self):
        with patch('subprocess.check_output', side_effect=self.output) as co:
            tests = command_line._get_test_ids_for_suites(self.suites)
            self.assertEqual(tests, self.tests)
            self.assertEqual(co.call_count, len(self.suites))
            count = 0
            for call in co.call_args_list:
                self.assertIn(self.suites[count], call[0][0])
                count += 1


class TestListAutopilotErrorTestCase(ConfigBaseTestCase):

    def test_autopilot_import_error_raises(self):
        ERROR = b'\nImportError: Could not import module.\n'
        PATCH = 'subprocess.check_output'
        with patch(PATCH, return_value=ERROR), self.assertRaises(ImportError):
            command_line._list_autopilot_tests('suite')

    def test_autopilot_suite_error_raises(self):
        PATCH = 'subprocess.check_output'
        EXCP = CalledProcessError(1, ['autopilot3', 'list', 'suite'])
        with patch(PATCH, side_effect=EXCP), self.assertRaises(ValueError):
            command_line._list_autopilot_tests('suite')


class TestIsValidTestIDTestCase(ConfigBaseTestCase,
                                testscenarios.TestWithScenarios):

    scenarios = [
        ('suite', {
            'id': 'suite1',
            'result': True,
        }),
        ('module', {
            'id': 'suite1.module1',
            'result': True,
        }),
        ('class', {
            'id': 'suite1.module1.Class1',
            'result': True,
        }),
        ('test', {
            'id': 'suite1.module1.Class1.test1',
            'result': True,
        }),
        ('characters and whitespace', {
            'id': 'ImportError: cannot import',
            'result': False,
        }),
        ('new line', {
            'id': '\n',
            'result': False,
        }),
        ('whitespace', {
            'id': ' ',
            'result': False,
        }),
    ]

    def test_valid_test_id(self):
        self.assertEqual(command_line._is_valid_test_id(self.id), self.result)


class TestSuitesParameterFormat(ConfigBaseTestCase,
                                testscenarios.TestWithScenarios):

    scenarios = [
        ('single suite', {
            'suites': ['suite1'],
            'result': ['suite1'],
        }),
        ('multi suite', {
            'suites': ['suite1', 'suite2'],
            'result': ['suite1', 'suite2'],
        }),
        ('space separated', {
            'suites': ['suite1 suite2'],
            'result': ['suite1', 'suite2'],
        }),
        ('leading trailing whitespace', {
            'suites': [' suite1 suite2 '],
            'result': ['suite1', 'suite2'],
        }),
        ('multi space separated', {
            'suites': ['suite1 suite2', 'suite3 suite4', 'suite5'],
            'result': ['suite1', 'suite2', 'suite3', 'suite4', 'suite5'],
        }),
    ]

    def test_suites_formatting(self):

        class SuitesArg():
            def __init__(self, suites):
                self.suites = suites

        args = SuitesArg(self.suites)
        command_line._validate_suites(args)
        self.assertEqual(args.suites, self.result)
