# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import testscenarios

from unittest import mock

from ubuntu_system_tests.host.command_line import (
    FIELD_EXTERNAL_ID,
    FIELD_VALUE,
)
from ubuntu_system_tests.host import practitest


class DummyResponse:
    """Class used to represent a dummy http response."""
    def __init__(self, json_resp=None, ok=True, text=None):
        self.json_resp = json_resp
        self.ok = ok
        self.text = text

    def json(self):
        return self.json_resp


@mock.patch(
    'ubuntu_system_tests.host.practitest.requests.get')
class PractitestSessionTestCase(testscenarios.TestWithScenarios):

    RESP_MULTI_TEST = DummyResponse(
        {'data': [
            {FIELD_EXTERNAL_ID: {FIELD_VALUE: 'module1.file1.class1.test1'}},
            {FIELD_EXTERNAL_ID: {FIELD_VALUE: 'module2.file1.class1.test1'}},
            {FIELD_EXTERNAL_ID: {FIELD_VALUE: 'module3.file1.class1.test1'}},
        ],
            'additional_data': {'pagination': {
                'limit': 250, 'page': 1, 'total_entities': 3}}
        }
    )

    RESP_SINGLE_TEST = DummyResponse(
        {FIELD_EXTERNAL_ID: {FIELD_VALUE: 'module1.file1.class1.test1'}}
    )

    RESP_ERROR = DummyResponse(ok=False, text='{"error": "msg"}')

    def test_no_auth_token_raises(self, mock_get):
        with self.assertRaises(ValueError):
            practitest.PractitestSession('')

    def test_request_header_contains_api_token(self, mock_get):
        pts = practitest.PractitestSession('123', '456')
        pts.get_test('1234')
        args, kwargs = mock_get.call_args
        self.assertEqual(
            kwargs['headers']['Authorization'], 'custom api_token=123')

    def test_request_contains_project_id(self, mock_get):
        pts = practitest.PractitestSession('123', '456')
        pts.get_test('1234')
        args, kwargs = mock_get.call_args
        self.assertEqual(kwargs['data'], '{"project_id": "456"}')

    def test_request_contains_no_project_id(self, mock_get):
        pts = practitest.PractitestSession('123')
        pts.get_test('1234')
        args, kwargs = mock_get.call_args
        self.assertFalse(kwargs['data'].find('project_id') >= 0)

    def test_get_test(self, mock_get):
        pts = practitest.PractitestSession('1234', '5678')
        mock_get.return_value = self.RESP_SINGLE_TEST
        pt_list = pts.get_test('1234')
        self.assertEqual(pt_list, self.RESP_SINGLE_TEST.json())

    def test_get_tests(self, mock_get):
        pts = practitest.PractitestSession('1234', '5678')
        mock_get.return_value = self.RESP_MULTI_TEST
        pt_list = pts.get_tests('1234')
        self.assertEqual(pt_list, self.RESP_MULTI_TEST.json()['data'])

    def test_error_raises(self, mock_get):
        mock_get.return_value = self.RESP_ERROR
        pts = practitest.PractitestSession('1234', '5678')
        with self.assertRaises(ValueError):
            pts.get_test('1234')
