# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import tempfile

import types

import fixtures
import testscenarios

from ubuntu_system_tests.common.utils import get_tests_ids_to_run
from ubuntu_system_tests.external_tests.utils import get_external_test_ids
from ubuntu_system_tests.selftests.utils import ConfigBaseTestCase


class ExternalTestsListFile(fixtures.Fixture):

    def __init__(self, tests):
        super().__init__()
        if isinstance(tests, list):
            self.contents = '\n'.join(tests)
        else:
            self.contents = tests

    def setUp(self):
        super().setUp()
        _, self.path = tempfile.mkstemp()
        self.addCleanup(os.remove, self.path)
        with open(self.path, 'w') as test_lists_file:
            test_lists_file.write(self.contents)


class GetExternalIdsTestsTestCase(testscenarios.TestWithScenarios):

    scenarios = [
        ('empty file', {
            'file_contents': '',
            'expected_test_ids': []
        }),
        ('file with spaces', {
            'file_contents': '   \n\n   \n  ',
            'expected_test_ids': []
        }),
        ('one test', {
            'file_contents': 'module.suite.test1',
            'expected_test_ids': ['module.suite.test1']
        }),
        ('multiple tests', {
            'file_contents': (
                'module.suite.test1\n'
                'module.suite.test2\n'
                'module.suite.test3\n'
            ),
            'expected_test_ids': [
                'module.suite.test1',
                'module.suite.test2',
                'module.suite.test3'
            ]
        }),
        ('tests with spaces', {
            'file_contents': (
                '\n'
                '   module.suite.test1\n'
                '\n'
                'module.suite.test2   \n'
                '   module.suite.test3   \n'
                '\n'
            ),
            'expected_test_ids': [
                'module.suite.test1',
                'module.suite.test2',
                'module.suite.test3'
            ]
        })
    ]

    def test_get_external_test_ids(self):
        with ExternalTestsListFile(self.file_contents) as test_lists_file:
            test_ids = get_external_test_ids(
                test_lists_file.path)
        self.assertEqual(self.expected_test_ids, test_ids)


class FakeJobHelpers:

    def __init__(self):
        self.name = 'unity8'

    jobs = {}
    # A dictionary to hold job names and their status.

    def is_running(self):
        return self.jobs.get(self.name, False)

    def start(self):
        self.jobs[self.name] = True

    def stop(self):
        self.jobs[self.name] = False

    def ensure_stopped(self):
        if self.is_running():
            self.stop()
        assert not self.is_running()

    def ensure_unity_running_and_greeter_hidden(self):
        greeter = FakeGreeterHelpers()
        if not self.is_running():
            self.start()
            greeter.wait_for_greeter()
        greeter.hide_greeter_with_dbus()


class FakeGreeterHelpers:

    active = True

    def wait_for_greeter(self):
        assert self.active

    def hide_greeter_with_dbus(self):
        self.active = False


class GetTestIdsToRunTestCase(ConfigBaseTestCase):

    def test_get_tests_with_no_internal_tests_module(self):
        ext_tests = ['external_test_1']
        with ExternalTestsListFile(ext_tests) as test_lists_file:
            tests_to_run = get_tests_ids_to_run(
                tests_module='not a module',
                testsuite_list_filepath=test_lists_file.path)

        self.assertEqual(ext_tests, tests_to_run)

    def test_get_tests_with_internal_tests_module(self):
        """Test that the internal tests module is always in the list to run."""
        ext_tests = ['external_test_1']
        tests_module = types.ModuleType('internal_tests_module')

        with ExternalTestsListFile(ext_tests) as test_lists_file:
            tests_to_run = get_tests_ids_to_run(
                tests_module, test_lists_file.path)

        expected_tests = ext_tests + ['internal_tests_module']
        self.assertEqual(expected_tests, tests_to_run)
