# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015, 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import shutil

from testtools import skipUnless

from ubuntu_system_tests.helpers.indicators.indicators import IndicatorsFactory
from ubuntu_system_tests.helpers.indicators.sound import SoundIndicatorPage

from ubuntu_system_tests.helpers import music as music_helper
from ubuntu_system_tests.helpers import scopes
from ubuntu_system_tests.helpers import wait_until
from ubuntu_system_tests.helpers.scopes import music

from ubuntu_system_tests.tests import data as test_data
from ubuntu_system_tests.tests.scopes import BaseScopeTestCase, SKIP_REASON


class MusicScopeTest(BaseScopeTestCase):
    """Tests for the Music scope."""

    def check_all_included(self, list1, list2):
        """ Verify all the elements in the list1 are includede in the list2 """
        for elem in list1:
            self.assertIn(elem, list2)

    def _copy_test_data_and_open_artist_preview(self):
        shutil.copy2(test_data.SRC_AUDIO_FILE_2, test_data.LOCAL_AUDIO_FILE_2)

        scope = scopes.go_to_music_scope()

        # Select the artist
        self.assertTrue(
            scope.is_artist(self.test_metadata.audio.count20.artist))
        return scope.select_artist(self.test_metadata.audio.count20.artist)

    @skipUnless(scopes.is_scope_included(scopes.SCOPE_ID_MUSIC), SKIP_REASON)
    def test_play_music_from_music_scope(self):
        """ Verify it is possible to reproduce music from the music scope """
        artist_preview = self._copy_test_data_and_open_artist_preview()

        # Play and pause the song and check the song is being played
        track_preview = artist_preview.select_track(
            self.test_metadata.audio.count20.title)
        track_preview.play()
        self.assertTrue(track_preview.is_progress_bar_moving(),
                        'Progress bar is not moving while the song in playing')
        track_preview.pause()
        self.assertFalse(track_preview.is_progress_bar_moving(),
                         'Progress bar is moving while the song in paused')

        # Verify the music app is playing the selected song
        track_preview.play_in_music_app()
        self.assertTrue(wait_until(music_helper.is_music_app_running))
        self.addCleanup(music_helper.stop_music_app)
        self.assertEqual(
            test_data.LOCAL_AUDIO_FILE_2, music_helper.get_current_song())

    @skipUnless(scopes.is_scope_included(scopes.SCOPE_ID_MUSIC), SKIP_REASON)
    def test_play_music_from_scope_pause_from_indicator(self):
        """
        Verify it is possible to play music from scope and pause from
        indicator.
        """
        artist_preview = self._copy_test_data_and_open_artist_preview()

        # Play and pause the song and check the song is being played
        track_preview = artist_preview.select_track(
            self.test_metadata.audio.count20.title)
        track_preview.play()
        self.assertTrue(track_preview.is_progress_bar_moving(),
                        'Progress bar is not moving while the song in playing')

        indicator_factory = IndicatorsFactory()
        sound_indicator = indicator_factory.get_indicators_page(
            SoundIndicatorPage.name)
        sound_indicator.playback_controls.pause_track()

        self.assertFalse(track_preview.is_progress_bar_moving(),
                         'Progress bar is moving while the song in playing')

    @skipUnless(scopes.is_scope_included(scopes.SCOPE_ID_MUSIC), SKIP_REASON)
    def test_music_scope_displays_local_music(self):
        """
        Verify music stored in internal storage of the device is shown in the
        music scope and the music artist and albums are correctly displayed in
        my music screen
        """
        shutil.copy2(test_data.SRC_AUDIO_FILE_1, test_data.LOCAL_AUDIO_FILE_1)
        shutil.copy2(test_data.SRC_AUDIO_FILE_2, test_data.LOCAL_AUDIO_FILE_2)
        shutil.copy2(test_data.SRC_AUDIO_FILE_3, test_data.LOCAL_AUDIO_FILE_3)
        songs = [self.test_metadata.audio.ogg_file,
                 self.test_metadata.audio.mp3_file,
                 self.test_metadata.audio.count20]

        scope = scopes.go_to_music_scope()

        # Validate all the artists displayed are the expected
        song_artists = music.get_artists(songs, remove_duplicate=False)
        displayed_artists = sorted(scope.get_subtitles(
            scope.CATEGORY_MY_MUSIC))
        self.assertEqual(song_artists, displayed_artists)

        artists_preview = scope.select_category(
            scope.HEADER_MY_MUSIC)

        # Validate all the artists displayed are the expected
        category_artist_section = scope.get_category_section(
            scope.CATEGORY_ARTISTS)
        artists = sorted(artists_preview.get_displayed_titles(
            category_artist_section))
        album_artists = music.get_artists(songs, remove_duplicate=True)
        self.assertEqual(album_artists, artists)

        # Validate all the albums displayed are the expected
        artist_preview = artists_preview.select_artist(song_artists[0])
        albums = sorted(artist_preview.get_displayed_titles())
        self.assertEqual(
            [self.test_metadata.audio.count20.album], albums)

    @skipUnless(scopes.is_scope_included(scopes.SCOPE_ID_MUSIC), SKIP_REASON)
    def test_music_scope_displays_online_music(self):
        """
        Verify the music scope show different online providers such as youtube
        and sound cloud. The songs that are shown in the music scope for those
        providers are also shown in their category previews
        """
        scope = scopes.go_to_music_scope()
        scope_data = scopes.get_scope_data(scopes.SCOPE_ID_MUSIC)

        # Get the expected sections ids based on the channel and
        # device being used
        ids = scope_data.get_expected_sections_ids()

        # Iterate over all the expected categories, select the category
        # header to access the category details and verify that the titles
        # display in the category section within the main view are also
        # displayed in the category details
        for category in scope.get_online_music_categories(ids):
            scope_titles = scope.get_titles(category['name'])
            scope_view = scope.select_category(category['header'])
            preview_titles = scope_view.get_displayed_titles()

            self.assertLessEqual(2, len(scope_titles))
            self.check_all_included(scope_titles,
                                    preview_titles)
            scopes.go_to_music_scope()

    @skipUnless(scopes.is_scope_included(scopes.SCOPE_ID_MUSIC), SKIP_REASON)
    def test_digital_category_search_by_top_albums_label(self):
        """""
        1. Access the digital category details,
        2. Clear the filtering options
        3. Filter the titles by Top Albums
        4. Verify that the displayed titles contains at least 2 titles
        5. Go back to the main view
        """
        scope = scopes.go_to_music_scope()
        scope_view = scope.select_category(
            scope.ONLINE_MUSIC_CATEGORIES['7digital']['header']
        )
        music_header = music.get_page_header()
        music_header.clear_search()
        music_header.select_search([music.TOP_ALBUMS_LABEL])

        self.assertLessEqual(2, len(
            scope_view.get_titles_list()))
        music_header.clear_search()
        self.dash.get_current_page_header().go_back()
