# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014-2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from autopilot.matchers import Eventually
from testtools.matchers import Equals

from ubuntu_system_tests.helpers.addressbook import AddressBook
from ubuntu_system_tests.helpers.camera import (
    Camera,
    app as camera_app_helpers,
    fixture_setup as camera_fixture)
from ubuntu_system_tests.helpers.dialer_app import (
    Dialer,
    app as dialer_app_helpers)
from ubuntu_system_tests.helpers.messaging import Messaging
from ubuntu_system_tests.helpers.processes import is_process_running
from ubuntu_system_tests.helpers.unity8.utils import APP_WIN_NAME

from ubuntu_system_tests.tests import base


class AppSwitcherTestCase(base.BaseUbuntuSystemTestCase):

    def setUp(self):
        super().setUp()
        self.useFixture(camera_fixture.SetCameraAccessRequests())

    def _switch_to_app(self, stage, app_id, app_name):
        """Select an app from the task switcher and check it is displayed."""
        stage.switch_to_app(app_name)
        self.assertThat(self.shell.get_current_focused_app,
                        Eventually(Equals(app_id)))

    def _close_app(self, stage, app_id, app_name):
        """Close an app using task switcher and check process is stopped."""
        self.assertTrue(is_process_running(app_id))
        app_windows = stage.get_app_window_names()
        count = len(app_windows)
        self.assertTrue(app_name in app_windows)
        stage.close_app(app_name)
        app_windows = stage.get_app_window_names()
        self.assertFalse(app_name in app_windows)
        self.assertFalse(is_process_running(app_id))
        self.assertEqual(len(app_windows), count - 1)

    def _launch_switcher(self):
        """ Open the task switcher and return it """
        stage = self.shell.swipe_to_show_app_switcher()
        stage.swipe_to_top()
        return stage

    def _launch_applications(self, stage):
        """Launch test applications and return their titles """
        app_classes = [Camera, Messaging, AddressBook, Dialer]
        app_titles = []
        for app_class in app_classes:
            app = app_class()
            app.launch_from_launcher(return_proxy=False)
            app_titles.append(app.get_main_view().get_main_title())
        apps = stage.get_app_window_names()
        expected_names = app_titles.copy()
        expected_names.append(APP_WIN_NAME)
        self.assertEqual(sorted(expected_names), sorted(apps))
        return app_titles

    def test_app_selection(self):
        """
        Launch applications and check that it is possible to switch focus
        back to them using the task switcher.
        """
        stage = self._launch_switcher()
        camera_title, ma, aa, dialer_title = self._launch_applications(stage)
        self._switch_to_app(stage, camera_app_helpers.APP_ID, camera_title)
        stage = self.shell.swipe_to_show_app_switcher()
        self._switch_to_app(stage, dialer_app_helpers.APP_WIN_ID, dialer_title)

    def test_close_app(self):
        """
        Launch applications and check it is possible to close them using
        the swipe actions.
        """
        stage = self._launch_switcher()
        dialer_title = self._launch_applications(stage)[3]
        self._close_app(stage, dialer_app_helpers.APP_WIN_ID, dialer_title)
